# Generated by Django 4.2.16 on 2024-11-09 21:23

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("afat", "0023_the_big_rename"),
    ]

    operations = [
        migrations.CreateModel(
            name="Doctrine",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Short name to identify this doctrine",
                        max_length=255,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="You can add notes about this doctrine here if you want. (optional)",
                        verbose_name="Notes",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="Whether this doctrine is enabled or not.",
                        verbose_name="Is enabled",
                    ),
                ),
            ],
            options={
                "verbose_name": "Doctrine",
                "verbose_name_plural": "Doctrines",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "default_fatlink_expiry_time",
                    models.PositiveIntegerField(
                        default=60,
                        help_text="Default expiry time for clickable FAT links in minutes. (Default: 60 minutes)",
                        verbose_name="Default FAT link expiry time",
                    ),
                ),
                (
                    "default_fatlink_reopen_grace_time",
                    models.PositiveIntegerField(
                        default=60,
                        help_text="Default time in minutes a FAT link can be re-opened after it is expired. (Default: 60 minutes)",
                        verbose_name="Default FAT link reopen grace time",
                    ),
                ),
                (
                    "default_fatlink_reopen_duration",
                    models.PositiveIntegerField(
                        default=60,
                        help_text="Default time in minutes a FAT link is re-opened for. (Default: 60 minutes)",
                        verbose_name="Default FAT link reopen duration",
                    ),
                ),
                (
                    "default_log_duration",
                    models.PositiveIntegerField(
                        default=60,
                        help_text="Default time in days a log entry is kept. (Default: 60 days)",
                        verbose_name="Default log duration",
                    ),
                ),
                (
                    "use_doctrines_from_fittings_module",
                    models.BooleanField(
                        db_index=True,
                        default=False,
                        help_text="Whether to use the doctrines from the Fittings modules in the doctrine dropdown. Note: The fittings module needs to be installed for this.",
                        verbose_name="Use doctrines from fittings module",
                    ),
                ),
            ],
            options={
                "verbose_name": "Setting",
                "verbose_name_plural": "Settings",
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="fatlink",
            name="doctrine",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The FAT link doctrine",
                max_length=254,
            ),
        ),
        migrations.AddField(
            model_name="fatlink",
            name="fleet_type",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The FAT link fleet type, if it's set",
                max_length=254,
            ),
        ),
        migrations.AlterField(
            model_name="fatlink",
            name="link_type",
            field=models.ForeignKey(
                help_text="Deprecated setting, will be removed in a future version …",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="afat.fleettype",
            ),
        ),
    ]



"""
Managers for our models
"""

# Django
from django.db import models
from django.db.models import Count, F


class FatLinkQuerySet(models.QuerySet):
    """
    FAT link queryset
    """

    def annotate_fats_count(self):
        """
        Annotate the amount fats per fat link

        :return:
        :rtype:
        """

        return self.annotate(fats_count=Count(expression=F("afat_fats")))


class FatLinkManager(models.Manager):
    """
    FAT link manager
    """

    def get_queryset(self) -> models.QuerySet:
        """
        Integrate custom QuerySet methods.
        """

        return FatLinkQuerySet(self.model, using=self._db)

    def select_related_default(self):
        """
        Apply select_related for default query optimizations.
        """

        return self.select_related(
            "creator", "character", "creator__profile__main_character"
        )


class FatManager(models.Manager):
    """
    FAT manager
    """

    def select_related_default(self):
        """
        Apply select_related for default query optimizations.
        """

        return self.select_related("fatlink", "character")



# Generated by Django 4.0.10 on 2023-04-21 21:58

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("afat", "0021_alter_afatlink_fleet"),
    ]

    operations = [
        migrations.AddField(
            model_name="afatlink",
            name="esi_error_count",
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name="afatlink",
            name="last_esi_error",
            field=models.CharField(
                blank=True,
                choices=[
                    (
                        "NOT_IN_FLEET",
                        "FC is not in the registered fleet anymore or fleet is no longer available.",
                    ),
                    ("NO_FLEET", "Registered fleet seems to be no longer available."),
                    ("NOT_FLEETBOSS", "FC is no longer the fleet boss."),
                ],
                default="",
                max_length=15,
            ),
        ),
        migrations.AddField(
            model_name="afatlink",
            name="last_esi_error_time",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="is_registered_on_esi",
            field=models.BooleanField(
                default=False,
                help_text="Whether the fleet to this FAT link is available in ESI or not",
            ),
        ),
    ]



"""
Test checks for installed modules we might use
"""

# Django
from django.test import override_settings

# Alliance Auth AFAT
from afat.app_settings import debug_enabled
from afat.tests import BaseTestCase


class TestDebugCheck(BaseTestCase):
    """
    Test if debug is enabled
    """

    @override_settings(DEBUG=True)
    def test_debug_enabled_with_debug_true(self) -> None:
        """
        Test debug_enabled with DEBUG = True

        :return:
        :rtype:
        """

        self.assertTrue(debug_enabled())

    @override_settings(DEBUG=False)
    def test_debug_enabled_with_debug_false(self) -> None:
        """
        Test debug_enabled with DEBUG = False

        :return:
        :rtype:
        """

        self.assertFalse(debug_enabled())



# Generated by Django 4.2.21 on 2025-05-18 18:20

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("afat", "0025_fat_alliance_eve_id_fat_corporation_eve_id"),
    ]

    operations = [
        migrations.CreateModel(
            name="FatsInTimeFilter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The filter name that is shown to the admin.",
                        max_length=500,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        help_text="The filter description that is shown to end users.",
                        max_length=500,
                    ),
                ),
                (
                    "days",
                    models.IntegerField(
                        default=30,
                        help_text="The number of days to look back for FATs.",
                    ),
                ),
                (
                    "fats_needed",
                    models.IntegerField(
                        default=10,
                        help_text="The number of FATs needed to pass the filter.",
                    ),
                ),
                (
                    "fleet_types",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Any of the selected fleet types are needed to pass the filter.",
                        to="afat.fleettype",
                    ),
                ),
                (
                    "ship_classes",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Any of the selected ship classes are needed to pass the filter.",
                        limit_choices_to={"eve_group__eve_category_id": 6},
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Smart Filter: FATs in time period",
                "verbose_name_plural": "Smart Filter: FATs in time period",
            },
        ),
    ]



"""
Test cases for the admin.py module.
"""

# Standard Library
from unittest.mock import MagicMock, patch

# Django
from django.contrib import admin

# Alliance Auth AFAT
from afat.admin import (
    AFatLinkTypeAdmin,
    AFatLogAdmin,
    DoctrineAdmin,
    FatsInTimeFilterAdmin,
    SettingAdmin,
)
from afat.forms import DoctrineAdminForm, SettingAdminForm
from afat.models import Doctrine, FatsInTimeFilter, FleetType, Log, Setting
from afat.tests import BaseTestCase


class TestAFatLinkTypeAdmin(BaseTestCase):
    """
    Test cases for the AFatLinkTypeAdmin class.
    """

    def test_activates_selected_fleet_types_successfully(self):
        """
        Test that the activate method successfully activates selected fleet types.

        :return:
        :rtype:
        """

        queryset = MagicMock()
        queryset.count.return_value = 3
        queryset.__iter__.return_value = [
            MagicMock(is_enabled=False),
            MagicMock(is_enabled=False),
            MagicMock(is_enabled=False),
        ]
        admin_instance = AFatLinkTypeAdmin(model=FleetType, admin_site=admin.site)
        admin_instance.activate(request=MagicMock(), queryset=queryset)

        for obj in queryset:
            self.assertTrue(obj.is_enabled)

    def test_deactivates_selected_fleet_types_successfully(self):
        """
        Test that the deactivate method successfully deactivates selected fleet types.

        :return:
        :rtype:
        """

        queryset = MagicMock()
        queryset.count.return_value = 3
        queryset.__iter__.return_value = [
            MagicMock(is_enabled=True),
            MagicMock(is_enabled=True),
            MagicMock(is_enabled=True),
        ]
        admin_instance = AFatLinkTypeAdmin(model=FleetType, admin_site=admin.site)
        admin_instance.deactivate(request=MagicMock(), queryset=queryset)

        for obj in queryset:
            self.assertFalse(obj.is_enabled)

    def test_handles_activation_failure_gracefully(self):
        """
        Test that the activate method handles activation failure gracefully.

        :return:
        :rtype:
        """

        queryset = MagicMock()
        queryset.count.return_value = 2
        mock_obj_1 = MagicMock(is_enabled=False)
        mock_obj_1.save.side_effect = Exception()  # Simulate save failure
        mock_obj_2 = MagicMock(is_enabled=False)
        queryset.__iter__.return_value = [mock_obj_1, mock_obj_2]

        admin_instance = AFatLinkTypeAdmin(model=FleetType, admin_site=admin.site)
        request = MagicMock()

        with patch.object(mock_obj_1, "is_enabled", new_callable=lambda: False):
            try:
                admin_instance.activate(request=request, queryset=queryset)
            except Exception:
                pass  # Ensure the test continues even if an exception is raised

        self.assertFalse(mock_obj_1.is_enabled)  # Ensure it remains False after failure
        self.assertTrue(mock_obj_2.is_enabled)  # Ensure successful activation

    def test_handles_deactivation_failure_gracefully(self):
        """
        Test that the deactivate method handles deactivation failure gracefully.

        :return:
        :rtype:
        """

        queryset = MagicMock()
        queryset.count.return_value = 2
        mock_obj_1 = MagicMock(is_enabled=True)
        mock_obj_1.save.side_effect = Exception()  # Simulate save failure
        mock_obj_2 = MagicMock(is_enabled=True)
        queryset.__iter__.return_value = [mock_obj_1, mock_obj_2]

        admin_instance = AFatLinkTypeAdmin(model=FleetType, admin_site=admin.site)
        request = MagicMock()

        with patch.object(mock_obj_1, "is_enabled", new_callable=lambda: True):
            try:
                admin_instance.deactivate(request=request, queryset=queryset)
            except Exception:
                pass  # Ensure the test continues even if an exception is raised

        self.assertTrue(mock_obj_1.is_enabled)  # Ensure it remains True after failure
        self.assertFalse(mock_obj_2.is_enabled)  # Ensure successful deactivation


class TestAFatLogAdmin(BaseTestCase):
    """
    Test cases for the AFatLogAdmin class.
    """

    def test_displays_all_log_fields_in_list_display(self):
        """
        Test that the list_display attribute contains all fields of the Log model.

        :return:
        :rtype:
        """

        MagicMock(
            log_time="2023-01-01 12:00:00",
            log_event="Event",
            user="User",
            fatlink_hash="hash123",
            log_text="Log text",
        )
        admin_instance = AFatLogAdmin(model=Log, admin_site=admin.site)

        self.assertEqual(
            admin_instance.list_display,
            ("log_time", "log_event", "user", "fatlink_hash", "log_text"),
        )

    def test_filters_logs_by_event(self):
        """
        Test that the list_filter attribute contains the log_event field.

        :return:
        :rtype:
        """

        with patch.object(
            AFatLogAdmin, "get_queryset", return_value=MagicMock()
        ) as mock_get_queryset:
            mock_request = MagicMock()
            admin_instance = AFatLogAdmin(model=Log, admin_site=admin.site)
            filtered_queryset = admin_instance.get_queryset(mock_request)

            self.assertEqual(filtered_queryset, mock_get_queryset.return_value)

    def test_searches_logs_by_fields(self):
        """
        Test that the search_fields attribute contains the correct fields for searching.

        :return:
        :rtype:
        """

        admin_instance = AFatLogAdmin(model=Log, admin_site=admin.site)

        self.assertEqual(
            admin_instance.search_fields,
            (
                "fatlink_hash",
                "user__profile__main_character__character_name",
                "user__username",
            ),
        )


class TestSettingAdmin(BaseTestCase):
    """
    Test cases for the SettingAdmin class.
    """

    def test_displays_correct_form_class(self):
        """
        Test that the SettingAdmin class displays the correct form class.

        :return:
        :rtype:
        """

        admin_instance = SettingAdmin(model=Setting, admin_site=admin.site)

        self.assertEqual(admin_instance.form, SettingAdminForm)

    def test_handles_missing_form_gracefully(self):
        with patch("afat.admin.SettingAdmin.form", None):
            admin_instance = SettingAdmin(model=Setting, admin_site=admin.site)

            self.assertIsNone(admin_instance.form)


class TestDoctrineAdmin(BaseTestCase):
    """
    Test cases for the DoctrineAdmin class.
    """

    def test_displays_correct_form_class(self):
        """
        Test that the DoctrineAdmin class displays the correct form class.

        :return:
        :rtype:
        """

        admin_instance = DoctrineAdmin(model=Doctrine, admin_site=admin.site)

        self.assertEqual(admin_instance.form, DoctrineAdminForm)

    def test_displays_all_fields_in_list_display(self):
        """
        Test that the list_display attribute contains all fields of the Doctrine model.

        :return:
        :rtype:
        """

        mock_field_1 = MagicMock()
        mock_field_1.name = "field_1"
        mock_field_2 = MagicMock()
        mock_field_2.name = "field_2"
        Doctrine._meta.get_fields = MagicMock(return_value=[mock_field_1, mock_field_2])

        with patch.object(
            DoctrineAdmin, "list_display", new_callable=lambda: ["field_1", "field_2"]
        ):
            admin_instance = DoctrineAdmin(model=Doctrine, admin_site=admin.site)

            self.assertEqual(admin_instance.list_display, ["field_1", "field_2"])

    def test_handles_empty_fields_gracefully(self):
        Doctrine._meta.get_fields = MagicMock(return_value=[])

        with patch.object(DoctrineAdmin, "list_display", new_callable=lambda: []):
            admin_instance = DoctrineAdmin(model=Doctrine, admin_site=admin.site)

            self.assertEqual(admin_instance.list_display, [])


class TestFatsInTimeFilterAdmin(BaseTestCase):
    """
    Test cases for the FatsInTimeFilterAdmin class.
    """

    def test_returns_correct_fleet_types_for_filter(self):
        """
        Test that the get_fleet_types method returns the correct fleet types.

        :return:
        :rtype:
        """

        mock_fleet_type_1 = MagicMock()
        mock_fleet_type_1.name = "FleetType1"
        mock_fleet_type_2 = MagicMock()
        mock_fleet_type_2.name = "FleetType2"
        obj = MagicMock(
            fleet_types=MagicMock(
                all=MagicMock(return_value=[mock_fleet_type_1, mock_fleet_type_2])
            )
        )
        admin_instance = FatsInTimeFilterAdmin(
            model=FatsInTimeFilter, admin_site=admin.site
        )
        result = admin_instance.get_fleet_types(obj)

        self.assertEqual(result, "FleetType1, FleetType2")

    def test_returns_empty_fleet_types_when_none_exist(self):
        """
        Test that the get_fleet_types method returns an empty string when no fleet types exist.

        :return:
        :rtype:
        """

        obj = MagicMock(fleet_types=MagicMock(all=MagicMock(return_value=[])))
        admin_instance = FatsInTimeFilterAdmin(
            model=FatsInTimeFilter, admin_site=admin.site
        )
        result = admin_instance.get_fleet_types(obj)

        self.assertEqual(result, "")

    def test_returns_correct_ship_classes_for_filter(self):
        """
        Test that the get_ship_classes method returns the correct ship classes.

        :return:
        :rtype:
        """

        mock_ship_class_1 = MagicMock()
        mock_ship_class_1.name = "ShipClass1"
        mock_ship_class_2 = MagicMock()
        mock_ship_class_2.name = "ShipClass2"
        obj = MagicMock(
            ship_classes=MagicMock(
                all=MagicMock(return_value=[mock_ship_class_1, mock_ship_class_2])
            )
        )
        admin_instance = FatsInTimeFilterAdmin(
            model=FatsInTimeFilter, admin_site=admin.site
        )
        result = admin_instance.get_ship_classes(obj)

        self.assertEqual(result, "ShipClass1, ShipClass2")

    def test_returns_empty_ship_classes_when_none_exist(self):
        """
        Test that the get_ship_classes method returns an empty string when no ship classes exist.

        :return:
        :rtype:
        """

        obj = MagicMock(ship_classes=MagicMock(all=MagicMock(return_value=[])))
        admin_instance = FatsInTimeFilterAdmin(
            model=FatsInTimeFilter, admin_site=admin.site
        )
        result = admin_instance.get_ship_classes(obj)

        self.assertEqual(result, "")



"""
URL patterns for the dashboard
"""

# Django
from django.urls import path

# Alliance Auth AFAT
from afat.views import dashboard

urls = [path(route="", view=dashboard.overview, name="dashboard")]



"""
Utilities
"""

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.utils.functional import lazy
from django.utils.html import format_html

# Alliance Auth
from allianceauth.authentication.admin import User
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.providers import esi

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)

# Format for output of datetime for this app
DATETIME_FORMAT = "%Y-%m-%d %H:%M"

format_html_lazy = lazy(format_html, str)


class NoDataError(Exception):
    """
    NoDataError
    """

    def __init__(self, msg):
        Exception.__init__(self, msg)


def write_log(request: WSGIRequest, log_event: str, fatlink_hash: str, log_text: str):
    """
    Write the log

    :param request:
    :type request:
    :param log_event:
    :type log_event:
    :param fatlink_hash:
    :type fatlink_hash:
    :param log_text:
    :type log_text:
    :return:
    :rtype:
    """

    # Alliance Auth AFAT
    from afat.models import Log  # pylint: disable=import-outside-toplevel

    afat_log = Log()
    afat_log.user = request.user
    afat_log.log_event = log_event
    afat_log.log_text = log_text
    afat_log.fatlink_hash = fatlink_hash
    afat_log.save()


def get_or_create_character(
    name: str = None, character_id: int = None
) -> EveCharacter | None:
    """
    This function takes a name or id of a character and checks
    to see if the character already exists.
    If the character does not already exist, it will create the
    character object, and if needed the corp/alliance objects as well.

    :param name:
    :type name:
    :param character_id:
    :type character_id:
    :return:
    :rtype:
    """

    eve_character = None

    if name:
        # If a name is passed to this function, we have to check it on ESI
        result = esi.client.Universe.post_universe_ids(names=[name]).results()

        if "characters" not in result or result["characters"] is None:
            return None

        character_id = result["characters"][0]["id"]
        eve_character = EveCharacter.objects.filter(character_id=character_id)
    elif character_id:
        # If an ID is passed to this function, we can just check the db for it.
        eve_character = EveCharacter.objects.filter(character_id=character_id)
    elif not name and not character_id:
        raise NoDataError(msg="No character name or character id provided.")

    if eve_character is not None and len(eve_character) == 0:
        # Create character
        character = EveCharacter.objects.create_character(character_id=character_id)
        character = EveCharacter.objects.get(pk=character.pk)

        logger.info(msg=f"EveCharacter Object created: {character.character_name}")

        # Create alliance and corporation info objects if not already exists for
        # future sanity
        if character.alliance_id is not None:
            # Create alliance and corporation info objects if not already exists
            if not EveAllianceInfo.objects.filter(
                alliance_id=character.alliance_id
            ).exists():
                EveAllianceInfo.objects.create_alliance(
                    alliance_id=character.alliance_id
                )
        else:
            # Create corporation info object if not already exists
            if not EveCorporationInfo.objects.filter(
                corporation_id=character.corporation_id
            ).exists():
                EveCorporationInfo.objects.create_corporation(
                    corp_id=character.corporation_id
                )
    else:
        character = eve_character[0]

    return character


def get_or_create_corporation_info(corporation_id: int) -> EveCorporationInfo:
    """
    Get or create corporation info

    :param corporation_id:
    :type corporation_id:
    :return:
    :rtype:
    """

    try:
        eve_corporation_info = EveCorporationInfo.objects.get(
            corporation_id=corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        eve_corporation_info = EveCorporationInfo.objects.create_corporation(
            corp_id=corporation_id
        )

        logger.info(
            msg=f"EveCorporationInfo object created: {eve_corporation_info.corporation_name}"
        )

    return eve_corporation_info


def get_or_create_alliance_info(alliance_id: int) -> EveAllianceInfo:
    """
    Get or create alliance info

    :param alliance_id:
    :type alliance_id:
    :return:
    :rtype:
    """

    try:
        eve_alliance_info = EveAllianceInfo.objects.get(alliance_id=alliance_id)
    except EveAllianceInfo.DoesNotExist:
        eve_alliance_info = EveAllianceInfo.objects.create_alliance(
            alliance_id=alliance_id
        )

        logger.info(
            msg=f"EveAllianceInfo object created: {eve_alliance_info.alliance_name}"
        )

    return eve_alliance_info


def get_main_character_from_user(user: User) -> str:
    """
    Get the main character from a user

    :param user:
    :type user:
    :return:
    :rtype:
    """

    user_main_character = user.username

    try:
        user_profile = user.profile
        user_main_character = user_profile.main_character.character_name
    except AttributeError:
        pass

    return user_main_character



"""
Admin pages configuration
"""

# Third Party
from solo.admin import SingletonModelAdmin

# Django
from django.contrib import admin, messages
from django.db.models import Count
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

# Alliance Auth AFAT
from afat.app_settings import securegroups_installed
from afat.forms import DoctrineAdminForm, SettingAdminForm
from afat.models import (
    Doctrine,
    Fat,
    FatLink,
    FatsInTimeFilter,
    FleetType,
    Log,
    Setting,
)


# Register your models here.
@admin.register(FatLink)
class AFatLinkAdmin(admin.ModelAdmin):
    """
    Config for the FAT link model
    """

    list_display = (
        "created",
        "creator",
        "fleet",
        "fleet_type",
        "is_esilink",
        "hash",
        "number_of_fats",
    )
    list_filter = ("is_esilink", "fleet_type")
    ordering = ("-created",)
    search_fields = (
        "fleet_type",
        "hash",
        "fleet",
        "creator__profile__main_character__character_name",
        "creator__username",
    )
    exclude = ("link_type", "character")

    def get_queryset(self, request):
        """
        Get the queryset

        :param request:
        :type request:
        :return:
        :rtype:
        """

        queryset = super().get_queryset(request)
        queryset = queryset.annotate(
            _number_of_fats=Count(expression="afat_fats", distinct=True)
        )

        return queryset

    @admin.display(ordering="_number_of_fats")
    def number_of_fats(self, obj):
        """
        Return the number of FATs per FAT link

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj._number_of_fats


@admin.register(Fat)
class AFatAdmin(admin.ModelAdmin):
    """
    Config for fat model
    """

    list_display = ("character", "system", "shiptype", "fatlink")
    list_filter = ("character", "system", "shiptype")
    ordering = ("-character",)
    search_fields = (
        "character__character_name",
        "system",
        "shiptype",
        "fatlink__fleet",
        "fatlink__hash",
    )


@admin.register(FleetType)
class AFatLinkTypeAdmin(admin.ModelAdmin):
    """
    Config for the FAT link type model
    """

    list_display = ("id", "_name", "_is_enabled")
    list_filter = ("is_enabled",)
    ordering = ("name",)

    @admin.display(description=_("Fleet type"), ordering="name")
    def _name(self, obj):
        """
        Rewrite name

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.name

    @admin.display(description=_("Is enabled"), boolean=True, ordering="is_enabled")
    def _is_enabled(self, obj):
        """
        Rewrite is_enabled

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.is_enabled

    actions = ("activate", "deactivate")

    @admin.action(description=_("Activate selected fleet types"))
    def activate(self, request, queryset):
        """
        Mark fleet type as active

        :param request:
        :type request:
        :param queryset:
        :type queryset:
        :return:
        :rtype:
        """

        notifications_count = 0
        failed = 0

        for obj in queryset:
            try:
                obj.is_enabled = True
                obj.save()

                notifications_count += 1
            except Exception:  # pylint: disable=broad-exception-caught
                failed += 1

        if failed:
            messages.error(
                request,
                ngettext(
                    singular="Failed to activate {failed} fleet type",
                    plural="Failed to activate {failed} fleet types",
                    number=failed,
                ).format(failed=failed),
            )

        if queryset.count() - failed > 0:
            messages.success(
                request,
                ngettext(
                    singular="Activated {notifications_count} fleet type",
                    plural="Activated {notifications_count} fleet types",
                    number=notifications_count,
                ).format(notifications_count=notifications_count),
            )

    @admin.action(description=_("Deactivate selected fleet types"))
    def deactivate(self, request, queryset):
        """
        Mark fleet type as inactive

        :param request:
        :type request:
        :param queryset:
        :type queryset:
        :return:
        :rtype:
        """

        notifications_count = 0
        failed = 0

        for obj in queryset:
            try:
                obj.is_enabled = False
                obj.save()

                notifications_count += 1
            except Exception:  # pylint: disable=broad-exception-caught
                failed += 1

        if failed:
            messages.error(
                request,
                ngettext(
                    singular="Failed to deactivate {failed} fleet type",
                    plural="Failed to deactivate {failed} fleet types",
                    number=failed,
                ).format(failed=failed),
            )

        if queryset.count() - failed > 0:
            messages.success(
                request,
                ngettext(
                    singular="Deactivated {notifications_count} fleet type",
                    plural="Deactivated {notifications_count} fleet types",
                    number=notifications_count,
                ).format(notifications_count=notifications_count),
            )


@admin.register(Log)
class AFatLogAdmin(admin.ModelAdmin):
    """
    Config for the admin log model
    """

    list_display = ("log_time", "log_event", "user", "fatlink_hash", "log_text")
    ordering = ("-log_time",)
    readonly_fields = ("log_time", "log_event", "user", "fatlink_hash", "log_text")
    list_filter = ("log_event",)
    search_fields = (
        "fatlink_hash",
        "user__profile__main_character__character_name",
        "user__username",
    )


@admin.register(Setting)
class SettingAdmin(SingletonModelAdmin):
    """
    Setting Admin
    """

    form = SettingAdminForm


@admin.register(Doctrine)
class DoctrineAdmin(admin.ModelAdmin):
    """
    Doctrine Admin
    """

    form = DoctrineAdminForm

    # Display all fields in the admin page
    list_display = [field.name for field in Doctrine._meta.get_fields()]


class FatsInTimeFilterAdmin(admin.ModelAdmin):
    """
    Config for the FATs in time filter model
    """

    list_display = (
        "name",
        "description",
        "days",
        "fats_needed",
        "get_fleet_types",
        "get_ship_classes",
    )
    filter_horizontal = (
        "fleet_types",
        "ship_classes",
    )

    select_related = True

    @admin.display(description=_("Fleet types"))
    def get_fleet_types(self, obj):
        """
        Get fleet types

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return ", ".join([fleet_type.name for fleet_type in obj.fleet_types.all()])

    @admin.display(description=_("Ship classes"))
    def get_ship_classes(self, obj):
        """
        Get ship classes

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return ", ".join([ship_class.name for ship_class in obj.ship_classes.all()])


if securegroups_installed():
    admin.site.register(FatsInTimeFilter, FatsInTimeFilterAdmin)



"""
URLs for the AFAT app
"""

# Django
from django.urls import include, path

# Alliance Auth AFAT
from afat.constants import INTERNAL_URL_PREFIX
from afat.urls import statistics  # pylint: disable=W0406 E0611
from afat.urls import ajax, dashboard, fatlinks, logs

app_name: str = "afat"

urlpatterns = [
    # Dashboard
    path(route="", view=include(dashboard.urls)),
    # Log urls
    path(route="logs/", view=include(logs.urls)),
    # FAT Links urls
    path(route="fatlinks/", view=include(fatlinks.urls)),
    # Statistics urls
    path(route="statistics/", view=include(statistics.urls)),
    # Ajax calls urls
    path(route=f"{INTERNAL_URL_PREFIX}/ajax/", view=include(ajax.urls)),
]



# Generated by Django 3.1.1 on 2020-10-04 00:04

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("afat", "0011_change_meta_options_on_manualafat_20201003_2222"),
    ]

    operations = [
        migrations.AddField(
            model_name="manualafat",
            name="created_at",
            field=models.DateTimeField(
                blank=True, help_text="Time this FAT has been added manually", null=True
            ),
        ),
    ]



"""
Initialization of the afat.models package.
"""

# flake8: noqa

# Alliance Auth AFAT
from afat.models.afat import (
    Doctrine,
    Duration,
    Fat,
    FatLink,
    FleetType,
    General,
    Log,
    Setting,
    get_hash_on_save,
    get_sentinel_user,
)
from afat.models.smart_filter import FatsInTimeFilter



"""
Test user agent header for ESI requests
"""

# Django
# from django.conf import settings
#
# from afat.tests import BaseTestCase
#
# # Alliance Auth
# from esi import __url__ as esi_url
# from esi import __version__ as esi_version
#
# # Alliance Auth AFAT
# from afat import __app_name_useragent__, __github_url__, __version__
# from afat.providers import esi
#
#
# class TestUserAgent(BaseTestCase):
#     """
#     Test the user agent header for ESI requests
#     """
#
#     def test_user_agent_header(self):
#         """
#         Test that the user agent header is set correctly for ESI requests.
#
#         :return:
#         :rtype:
#         """
#
#         operation = esi.client.Universe.get_universe_factions()
#
#         self.assertEqual(
#             first=operation.future.request.headers["User-Agent"],
#             second=(
#                 f"{__app_name_useragent__}/{__version__} "
#                 f"({settings.ESI_USER_CONTACT_EMAIL}; +{__github_url__}) "
#                 f"Django-ESI/{esi_version} (+{esi_url})"
#             ),
#         )



# flake8: noqa

"""
Script to generate many fat links for load testing.

This script can be executed directly from shell.
"""


# Standard Library
import os
import sys
from datetime import timedelta
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

# Django
import django

# init and setup django project
os.environ.setdefault(key="DJANGO_SETTINGS_MODULE", value="myauth.settings.local")
django.setup()

# Standard Library
import random

# Django
from django.contrib.auth.models import User
from django.utils.timezone import now

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.helpers import random_string

# Alliance Auth AFAT
from afat.models import Fat, FatLink, Log
from afat.tests.fixtures.utils import RequestStub
from afat.utils import write_log

LINKS_NUMBER = 100
MAX_PILOTS_IN_FLEET = 256


characters = list(EveCharacter.objects.all())

print(
    f"Adding {LINKS_NUMBER:,} FAT links with up to {MAX_PILOTS_IN_FLEET} characters each"
)

user = User.objects.first()
creator = user.profile.main_character
fleet_type = "Generated Fleet"

for _ in range(LINKS_NUMBER):
    fat_link = FatLink.objects.create(
        fleet=f"Generated Fleet #{random.randint(a=1, b=1000000000)}",
        hash=random_string(char_count=20),
        creator=user,
        character=creator,
        fleet_type=fleet_type,
        created=now() - timedelta(days=random.randint(a=0, b=365)),
    )

    write_log(
        request=RequestStub(user=user),
        log_event=Log.Event.CREATE_FATLINK,
        log_text=(
            f'FAT link with name "{fat_link.fleet}" '
            f'(Fleet type: "{fleet_type}") was created'
        ),
        fatlink_hash=fat_link.hash,
    )

    for character in random.sample(
        population=characters, k=random.randint(a=1, b=MAX_PILOTS_IN_FLEET)
    ):
        Fat.objects.create(
            character_id=character.id,
            fatlink=fat_link,
            system="Jita",
            shiptype="Ibis",
        )

    print(".", end="", flush=True)


print("")
print("DONE")



"""
Test cases for the smart filter model.
"""

# Standard Library
from datetime import datetime, timedelta, timezone
from unittest.mock import MagicMock, patch

# Django
from django.contrib.auth.models import User
from django.db.models.signals import pre_save

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.authentication.signals import assign_state_on_active_change

# Alliance Auth AFAT
from afat.models.smart_filter import FatsInTimeFilter, _get_threshold_date
from afat.tests import BaseTestCase


class TestGetThresholdDate(BaseTestCase):
    """
    Test cases for the _get_threshold_date function.
    """

    def test_returns_correct_threshold_date_for_positive_days(self):
        """
        Test that the function returns the correct threshold date for positive days.

        :return:
        :rtype:
        """

        days = 10
        expected_date = datetime.now(timezone.utc) - timedelta(days=days)
        result = _get_threshold_date(days)

        self.assertAlmostEqual(result, expected_date, delta=timedelta(seconds=1))

    def test_returns_correct_threshold_date_for_zero_days(self):
        """
        Test that the function returns the correct threshold date for zero days.

        :return:
        :rtype:
        """

        days = 0
        expected_date = datetime.now(timezone.utc)
        result = _get_threshold_date(days)

        self.assertAlmostEqual(result, expected_date, delta=timedelta(seconds=1))


class TestFatsInTimeFilter(BaseTestCase):
    """
    Test cases for the FatsInTimeFilter class.
    """

    @patch("afat.models.smart_filter._get_threshold_date")
    @patch("afat.models.smart_filter.Fat.objects.filter")
    def test_returns_true_when_user_meets_fats_needed(
        self, mock_fat_filter, mock_get_threshold_date
    ):
        """
        Test that the process_filter method returns True when the user meets the fats needed.

        :param mock_fat_filter:
        :type mock_fat_filter:
        :param mock_get_threshold_date:
        :type mock_get_threshold_date:
        :return:
        :rtype:
        """

        mock_get_threshold_date.return_value = datetime.now(timezone.utc) - timedelta(
            days=30
        )
        mock_fat_filter.return_value.count.return_value = 10
        user = User.objects.create(username="testuser")  # Save the user to the database
        filter_instance = FatsInTimeFilter(days=30, fats_needed=10)
        filter_instance.save()  # Save the filter instance to the database

        self.assertTrue(filter_instance.process_filter(user))

    @patch("afat.models.smart_filter._get_threshold_date")
    @patch("afat.models.smart_filter.Fat.objects.filter")
    def test_returns_false_when_user_does_not_meet_fats_needed(
        self, mock_fat_filter, mock_get_threshold_date
    ):
        """
        Test that the process_filter method returns False when the user does not meet the fats needed.

        :param mock_fat_filter:
        :type mock_fat_filter:
        :param mock_get_threshold_date:
        :type mock_get_threshold_date:
        :return:
        :rtype:
        """

        mock_get_threshold_date.return_value = datetime.now(timezone.utc) - timedelta(
            days=30
        )
        mock_fat_filter.return_value.count.return_value = 5
        user = User.objects.create(username="testuser")  # Save the user to the database
        filter_instance = FatsInTimeFilter(days=30, fats_needed=10)
        filter_instance.save()  # Save the filter instance to the database

        self.assertFalse(filter_instance.process_filter(user))

    @patch("afat.models.smart_filter._get_threshold_date")
    @patch("afat.models.smart_filter.Fat.objects.filter")
    def test_returns_false_when_user_has_no_characters(
        self, mock_fat_filter, mock_get_threshold_date
    ):
        """
        Test that the process_filter method returns False when the user has no characters.

        :param mock_fat_filter:
        :type mock_fat_filter:
        :param mock_get_threshold_date:
        :type mock_get_threshold_date:
        :return:
        :rtype:
        """

        mock_get_threshold_date.return_value = datetime.now(timezone.utc) - timedelta(
            days=30
        )
        mock_fat_filter.side_effect = CharacterOwnership.DoesNotExist
        user = User.objects.create(username="testuser")  # Save the user to the database
        filter_instance = FatsInTimeFilter(days=30, fats_needed=10)
        filter_instance.save()  # Save the filter instance to the database

        self.assertFalse(filter_instance.process_filter(user))

    @patch("afat.models.smart_filter._get_threshold_date")
    @patch("afat.models.smart_filter.Fat.objects.filter")
    def test_returns_correct_audit_results_when_users_have_no_fats(
        self, mock_fat_filter, mock_get_threshold_date
    ):
        """
        Test that the audit_filter method returns the correct results when users have no fats.

        :param mock_fat_filter:
        :type mock_fat_filter:
        :param mock_get_threshold_date:
        :type mock_get_threshold_date:
        :return:
        :rtype:
        """

        mock_get_threshold_date.return_value = datetime.now(timezone.utc) - timedelta(
            days=30
        )

        mock_queryset = MagicMock()
        mock_queryset.select_related.return_value = mock_queryset
        mock_queryset.filter.return_value = mock_queryset
        mock_queryset.values_list.return_value = []
        mock_fat_filter.return_value = mock_queryset

        # Disconnect the signal to avoid triggering it during user creation
        pre_save.disconnect(assign_state_on_active_change, sender=User)

        user1 = User.objects.create(pk=1, username="user1")
        user2 = User.objects.create(pk=2, username="user2")

        # Reconnect the signal after user creation
        pre_save.connect(assign_state_on_active_change, sender=User)

        users = [user1, user2]
        filter_instance = FatsInTimeFilter(days=30, fats_needed=2)
        filter_instance.save()
        result = filter_instance.audit_filter(users)

        self.assertEqual(result[1]["message"], 0)
        self.assertFalse(result[1]["check"])
        self.assertEqual(result[2]["message"], 0)
        self.assertFalse(result[2]["check"])



"""
Migration script to fix hstats affiliation in statistics.
"""

# Django
from django.core.management import BaseCommand

# Alliance Auth AFAT
from afat.models import Fat
from afat.providers import esi


def get_input(text) -> str:
    """
    Wrapped input to enable import

    :param text:
    :type text:
    :return:
    :rtype:
    """

    return input(text)


class Command(BaseCommand):
    """
    Fix affiliation in statistics
    """

    def _get_character_corp_history(self, character_id: int) -> list:
        """
        Get character corporation history

        :param character_id:
        :type character_id:
        :return:
        :rtype:
        """

        self.stdout.write(
            msg=f"Querying ESI for corporation history for character with ID {character_id} …"
        )

        return esi.client.Character.get_characters_character_id_corporationhistory(
            character_id=character_id
        ).results()

    def _get_corporation_alliance_history(self, corporation_id: int) -> list:
        """
        Get corporation alliance history

        :param corporation_id:
        :type corporation_id:
        :return:
        :rtype:
        """

        self.stdout.write(
            msg=f"Querying ESI for alliance history for corporation with ID {corporation_id} …"
        )

        return esi.client.Corporation.get_corporations_corporation_id_alliancehistory(
            corporation_id=corporation_id
        ).results()

    def _fix_affiliation_in_statistics(self) -> None:
        """
        Fix affiliation in statistics

        :return:
        :rtype:
        """

        all_fats = Fat.objects.all().order_by("character_id")
        fats_total = all_fats.count()

        cache_character_corp_history = {}
        cache_corp_alliance_history = {}
        update_rows = []

        # Fix affiliation in statistics
        for loop_count, fat in enumerate(all_fats):
            self.stdout.write(
                msg=f"Migrating affiliation data for FAT ID {fat.pk} ({loop_count + 1}/{fats_total}) …"
            )

            if fat.character.character_id not in cache_character_corp_history:
                corp_history = self._get_character_corp_history(
                    character_id=fat.character.character_id
                )

                cache_character_corp_history[fat.character.character_id] = corp_history

            # print("cache_character_corp_history", cache_character_corp_history)

            for corp_entry in cache_character_corp_history[fat.character.character_id]:
                if fat.fatlink.created <= corp_entry["start_date"]:
                    continue

                affiliation_corp_id = corp_entry["corporation_id"]

                if affiliation_corp_id not in cache_corp_alliance_history:
                    alliance_history = self._get_corporation_alliance_history(
                        corporation_id=affiliation_corp_id
                    )

                    cache_corp_alliance_history[affiliation_corp_id] = alliance_history

                affiliation_alliance_id = next(
                    (
                        entry["alliance_id"]
                        for entry in cache_corp_alliance_history[affiliation_corp_id]
                        if entry.get("alliance_id")
                        and fat.fatlink.created > entry["start_date"]
                    ),
                    None,
                )

                update_rows.append(
                    {
                        "fat_id": fat.pk,
                        "corporation_eve_id": affiliation_corp_id,
                        "alliance_eve_id": affiliation_alliance_id,
                    }
                )
                break

        # Update rows
        Fat.objects.bulk_update(
            objs=[
                Fat(
                    pk=row["fat_id"],
                    corporation_eve_id=row["corporation_eve_id"],
                    alliance_eve_id=row["alliance_eve_id"],
                )
                for row in update_rows
            ],
            fields=["corporation_eve_id", "alliance_eve_id"],
            batch_size=500,
        )

        self.stdout.write(msg=self.style.SUCCESS("Migration complete!"))

    def handle(self, *args, **options):  # pylint: disable=unused-argument
        """
        Ask before running …

        :param args:
        :param options:
        :return:
        :rtype:
        """

        self.stdout.write(
            msg=self.style.SUCCESS(
                "This command will add the affiliation of pilots in all FATs so the "
                "statistics are correct. "
                "This might take quite a while depending on the number of pilots "
                "in your database. Please be patient."
            )
        )

        user_input = get_input(text="Are you sure you want to proceed? (yes/no) ")

        if user_input == "yes":
            self.stdout.write(msg="Starting migration. Please stand by.")
            self._fix_affiliation_in_statistics()
        else:
            self.stdout.write(msg=self.style.WARNING("Aborted."))



"""
Fatlinks URLs
These URLs are prefixed with `fatlinks/` and are used for FAT links management
"""

# Django
from django.urls import path

# Alliance Auth AFAT
from afat.views import fatlinks

urls = [
    # Fat links list actions
    path(
        route="",
        view=fatlinks.overview,
        name="fatlinks_overview",
    ),
    path(
        route="<int:year>/",
        view=fatlinks.overview,
        name="fatlinks_overview",
    ),
    # Fat link actions
    path(
        route="add/",
        view=fatlinks.add_fatlink,
        name="fatlinks_add_fatlink",
    ),
    path(
        route="link/create/esi-fatlink/",
        view=fatlinks.create_esi_fatlink,
        name="fatlinks_create_esi_fatlink",
    ),
    path(
        route="link/create/esi-fatlink/callback/<str:fatlink_hash>/",
        view=fatlinks.create_esi_fatlink_callback,
        name="fatlinks_create_esi_fatlink_callback",
    ),
    path(
        route="link/create/clickable-fatlink/",
        view=fatlinks.create_clickable_fatlink,
        name="fatlinks_create_clickable_fatlink",
    ),
    path(
        route="<str:fatlink_hash>/details/",
        view=fatlinks.details_fatlink,
        name="fatlinks_details_fatlink",
    ),
    path(
        route="<str:fatlink_hash>/delete/",
        view=fatlinks.delete_fatlink,
        name="fatlinks_delete_fatlink",
    ),
    path(
        route="<str:fatlink_hash>/stop-esi-tracking/",
        view=fatlinks.close_esi_fatlink,
        name="fatlinks_close_esi_fatlink",
    ),
    path(
        route="<str:fatlink_hash>/re-open/",
        view=fatlinks.reopen_fatlink,
        name="fatlinks_reopen_fatlink",
    ),
    # Fat actions
    path(
        route="<str:fatlink_hash>/register/",
        view=fatlinks.add_fat,
        name="fatlinks_add_fat",
    ),
    path(
        route="<str:fatlink_hash>/fat/<int:fat_id>/delete/",
        view=fatlinks.delete_fat,
        name="fatlinks_delete_fat",
    ),
]



"""
URLs for AJAX calls
These URLs are used for AJAX calls and are prefixed with `{INTERNAL_URL_PREFIX}/ajax/`
"""

# Django
from django.urls import path

# Alliance Auth AFAT
from afat.views import dashboard, fatlinks, logs, statistics

urls = [
    # Ajax calls :: Dashboard
    path(
        route="dashboard/get-recent-fatlinks/",
        view=dashboard.ajax_get_recent_fatlinks,
        name="dashboard_ajax_get_recent_fatlinks",
    ),
    path(
        route="dashboard/get-recent-fats-by-character/<int:charid>/",
        view=dashboard.ajax_recent_get_fats_by_character,
        name="dashboard_ajax_get_recent_fats_by_character",
    ),
    # Ajax calls :: Fat links
    path(
        route="fatlinks/get-fatlinks-by-year/<int:year>/",
        view=fatlinks.ajax_get_fatlinks_by_year,
        name="fatlinks_ajax_get_fatlinks_by_year",
    ),
    path(
        route="fatlinks/get-fats-by-fatlink/<str:fatlink_hash>/",
        view=fatlinks.ajax_get_fats_by_fatlink,
        name="fatlinks_ajax_get_fats_by_fatlink",
    ),
    # Ajax calls :: Logs
    path(
        route="logs/",
        view=logs.ajax_get_logs,
        name="logs_ajax_get_logs",
    ),
    # Ajax calls :: Statistics
    path(
        route="statistics/get-monthly-fats-for-main-character/<int:corporation_id>/<int:character_id>/<int:year>/<int:month>/",
        view=statistics.ajax_get_monthly_fats_for_main_character,
        name="statistics_ajax_get_monthly_fats_for_main_character",
    ),
]



"""
App config
"""

# Django
from django.utils.translation import gettext_lazy as _

__version__ = "3.12.3"
__title__ = _("Fleet Activity Tracking")

__package_name__ = "allianceauth-afat"
__app_name__ = "afat"
__app_name_verbose__ = "Alliance Auth AFAT"
__app_name_useragent__ = "Alliance-Auth-AFAT"

__github_url__ = f"https://github.com/ppfeufer/{__package_name__}"



"""
Helper for time related functions
"""

# Django
from django.utils.datetime_safe import datetime
from django.utils.translation import gettext as _


def get_time_delta(then, now=datetime.now(), interval="default"):
    """
    Returns a duration as specified by variable interval
    functions, except total_duration, returns [quotient, remainder]

    :param then:
    :type then:
    :param now:
    :type now:
    :param interval:
    :type interval:
    :return:
    :rtype:
    """

    duration = now.replace(tzinfo=None) - then.replace(tzinfo=None)
    duration_in_seconds = duration.total_seconds()

    def years():
        """
        Return years

        :return:
        :rtype:
        """

        return divmod(duration_in_seconds, 31536000)  # Seconds in a year = 31536000.

    def days(from_seconds=None):
        """
        Return days

        :param from_seconds:
        :type from_seconds:
        :return:
        :rtype:
        """

        return divmod(
            from_seconds if from_seconds is not None else duration_in_seconds,
            86400,  # Seconds in a day = 86400
        )

    def hours(from_seconds=None):
        """
        Return hours

        :param from_seconds:
        :type from_seconds:
        :return:
        :rtype:
        """

        return divmod(
            from_seconds if from_seconds is not None else duration_in_seconds,
            3600,  # Seconds in an hour = 3600
        )

    def minutes(from_seconds=None):
        """
        Return minutes

        :param from_seconds:
        :type from_seconds:
        :return:
        :rtype:
        """

        return divmod(
            from_seconds if from_seconds is not None else duration_in_seconds,
            60,  # Seconds in a minute = 60
        )

    def seconds(from_seconds=None):
        """
        Return seconds

        :param from_seconds:
        :type from_seconds:
        :return:
        :rtype:
        """

        if from_seconds is not None:
            return divmod(from_seconds, 1)

        return duration_in_seconds

    def total_duration():
        """
        Return total time difference

        :return:
        :rtype:
        """

        duration_years = years()
        duration_days = days(from_seconds=duration_years[1])
        duration_hours = hours(from_seconds=duration_days[1])
        duration_minutes = minutes(from_seconds=duration_hours[1])
        duration_seconds = seconds(from_seconds=duration_minutes[1])

        return _(
            f"{int(duration_years[0])} years, {int(duration_days[0])} days, {int(duration_hours[0])} hours, {int(duration_minutes[0])} minutes and {int(duration_seconds[0])} seconds"  # pylint: disable=line-too-long
        )

    return {
        "years": int(years()[0]),
        "days": int(days()[0]),
        "hours": int(hours()[0]),
        "minutes": int(minutes()[0]),
        "seconds": int(seconds()),
        "default": total_duration(),
    }[interval]



"""
Test fatlinks views
"""

# Standard Library
from http import HTTPStatus
from unittest.mock import ANY, MagicMock, patch

# Third Party
from pytz import utc

# Django
from django.contrib.messages import get_messages
from django.urls import reverse
from django.utils.datetime_safe import datetime

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# Alliance Auth AFAT
from afat.models import Duration, Fat, FatLink, Log, get_hash_on_save
from afat.tests import BaseTestCase
from afat.tests.fixtures.load_allianceauth import load_allianceauth
from afat.utils import get_main_character_from_user

MODULE_PATH = "afat.views.fatlinks"


class TestFatlinksView(BaseTestCase):
    """
    Test the fatlinks views
    """

    @classmethod
    def setUpClass(cls):
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()
        load_allianceauth()

        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_without_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1002.character_id,
            permissions=["afat.basic_access"],
        )

        cls.user_with_manage_afat, _ = create_user_from_evecharacter(
            character_id=cls.character_1003.character_id,
            permissions=["afat.basic_access", "afat.manage_afat"],
        )

        cls.user_with_add_fatlink, _ = create_user_from_evecharacter(
            character_id=cls.character_1004.character_id,
            permissions=["afat.basic_access", "afat.add_fatlink"],
        )

        # Generate some FAT links and FATs
        cls.afat_link_april_1 = FatLink.objects.create(
            fleet="April Fleet 1",
            hash="1231",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=1, tzinfo=utc),
        )
        cls.afat_link_april_2 = FatLink.objects.create(
            fleet="April Fleet 2",
            hash="1232",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=15, tzinfo=utc),
        )
        cls.afat_link_september = FatLink.objects.create(
            fleet="September Fleet",
            hash="1233",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=9, day=1, tzinfo=utc),
        )
        cls.afat_link_september_no_fats = FatLink.objects.create(
            fleet="September Fleet 2",
            hash="1234",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=9, day=1, tzinfo=utc),
        )

        Fat.objects.create(
            character=cls.character_1101,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1001,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1002,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1003,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1004,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1005,
            fatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )

        Fat.objects.create(
            character=cls.character_1001,
            fatlink=cls.afat_link_april_2,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1004,
            fatlink=cls.afat_link_april_2,
            shiptype="Thorax",
        )
        Fat.objects.create(
            character=cls.character_1002,
            fatlink=cls.afat_link_april_2,
            shiptype="Thorax",
        )
        Fat.objects.create(
            character=cls.character_1003,
            fatlink=cls.afat_link_april_2,
            shiptype="Omen",
        )

        Fat.objects.create(
            character=cls.character_1001,
            fatlink=cls.afat_link_september,
            shiptype="Omen",
        )
        Fat.objects.create(
            character=cls.character_1004,
            fatlink=cls.afat_link_september,
            shiptype="Guardian",
        )
        Fat.objects.create(
            character=cls.character_1002,
            fatlink=cls.afat_link_september,
            shiptype="Omen",
        )

    def test_should_show_fatlnks_overview(self):
        """
        Test should show fatlnks overview

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:fatlinks_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_fatlnks_overview_with_year(self):
        """
        Test should show fatlnks overview with year

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:fatlinks_overview", kwargs={"year": 2020})
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_add_fatlink_for_user_with_manage_afat(self):
        """
        Test should show add fatlink for user with manage afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(viewname="afat:fatlinks_add_fatlink")
        res = self.client.get(url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_add_fatlink_for_user_with_add_fatlinkt(self):
        """
        Test should show add fatlink for user with add fatlinkt

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_add_fatlink)

        url = reverse(viewname="afat:fatlinks_add_fatlink")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_fatlink_details_for_user_with_manage_afat(self):
        """
        Test should show fatlink details for user with manage afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": self.afat_link_april_1.hash},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_fatlink_details_for_user_with_add_fatlinkt(self):
        """
        Test should show fatlink details for user with add fatlinkt

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_add_fatlink)

        url = reverse(
            viewname="afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": self.afat_link_april_1.hash},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_fatlink_details_for_non_existing_fatlink(self):
        """
        Test should not show fatlink details for non existing fatlink

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": "foobarsson"},
        )
        res = self.client.get(path=url)

        self.assertNotEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

        messages = list(get_messages(request=res.wsgi_request))

        self.assertRaises(expected_exception=FatLink.DoesNotExist)
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second="<h4>Warning!</h4><p>The hash provided is not valid.</p>",
        )

    def test_ajax_get_fatlinks_by_year(self):
        """
        Test ajax get fatlinks by year

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        fatlink_hash = get_hash_on_save()
        fatlink_created = FatLink.objects.create(
            fleet="April Fleet 1",
            creator=self.user_with_manage_afat,
            character=self.character_1001,
            hash=fatlink_hash,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id=3726458287,
            fleet_type="CTA",
            doctrine="Ships",
            created="2021-11-05T13:19:49.676Z",
        )

        Duration.objects.create(fleet=fatlink_created, duration=120)

        fatlink = (
            FatLink.objects.select_related_default()
            .annotate_fats_count()
            .get(hash=fatlink_hash)
        )

        url_with_year = reverse(
            viewname="afat:fatlinks_ajax_get_fatlinks_by_year",
            kwargs={"year": 2021},
        )
        result = self.client.get(path=url_with_year)

        self.assertEqual(first=result.status_code, second=HTTPStatus.OK)

        creator_main_character = get_main_character_from_user(user=fatlink.creator)
        fleet_time = fatlink.created
        fleet_time_timestamp = fleet_time.timestamp()
        esi_marker = '<span class="badge text-bg-success afat-label ms-2">ESI</span>'

        close_esi_tracking_url = reverse(
            viewname="afat:fatlinks_close_esi_fatlink", args=[fatlink_hash]
        )
        redirect_url = reverse(viewname="afat:fatlinks_overview")
        edit_url = reverse(
            viewname="afat:fatlinks_details_fatlink", args=[fatlink_hash]
        )
        delete_url = reverse(
            viewname="afat:fatlinks_delete_fatlink", args=[fatlink_hash]
        )

        self.assertJSONEqual(
            raw=str(result.content, encoding="utf8"),
            expected_data=[
                {
                    "pk": fatlink.pk,
                    "fleet_name": fatlink.fleet + esi_marker,
                    "creator_name": creator_main_character,
                    "fleet_type": "CTA",
                    "fleet_time": {
                        "time": "2021-11-05T13:19:49.676Z",
                        "timestamp": fleet_time_timestamp,
                    },
                    "fats_number": 0,
                    "hash": fatlink.hash,
                    "is_esilink": True,
                    "doctrine": "Ships",
                    "esi_fleet_id": fatlink.esi_fleet_id,
                    "is_registered_on_esi": True,
                    "actions": (
                        '<a class="btn btn-afat-action btn-primary btn-sm" '
                        'style="margin-left: 0.25rem;" title="Clicking here will stop '
                        "the automatic tracking through ESI for this fleet and close "
                        'the associated FAT link." data-bs-toggle="modal" '
                        'data-bs-target="#cancelEsiFleetModal" '
                        f'data-url="{close_esi_tracking_url}?next={redirect_url}" '
                        'data-body-text="<p>Are you sure you want to close ESI fleet '
                        'with ID 3726458287 from Bruce Wayne?</p>" '
                        'data-confirm-text="Stop tracking"><i class="fa-solid fa-times">'
                        '</i></a><a class="btn btn-info btn-sm m-1" '
                        f'href="{edit_url}">'
                        '<span class="fa-solid fa-eye"></span></a>'
                        '<a class="btn btn-danger btn-sm" data-bs-toggle="modal" '
                        'data-bs-target="#deleteFatLinkModal" '
                        f'data-url="{delete_url}" '
                        'data-confirm-text="Delete" '
                        'data-body-text="<p>Are you sure you want to delete FAT '
                        'link April Fleet 1?</p>">'
                        '<i class="fa-solid fa-trash-can fa-fw"></i></a>'
                    ),
                    # "actions": "",
                    "via_esi": "Yes",
                }
            ],
        )

    @patch("afat.views.fatlinks.Duration.objects.get")
    @patch("afat.views.fatlinks.Setting.get_setting")
    @patch("afat.views.fatlinks.write_log")
    def test_reopens_fatlink_successfully(
        self, mock_write_log, mock_get_setting, mock_duration_get
    ):
        """
        Test reopens fatlink successfully

        :param mock_write_log:
        :type mock_write_log:
        :param mock_get_setting:
        :type mock_get_setting:
        :param mock_duration_get:
        :type mock_duration_get:
        :return:
        :rtype:
        """

        mock_duration = MagicMock()
        mock_duration.fleet.reopened = False
        mock_duration.fleet.created = datetime(2023, 1, 1, 12, 0, 0)
        mock_duration.fleet.hash = "test_hash"
        mock_get_setting.return_value = 60
        mock_duration_get.return_value = mock_duration

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_reopen_fatlink", args=["test_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            reverse(
                "afat:fatlinks_details_fatlink", kwargs={"fatlink_hash": "test_hash"}
            ),
            target_status_code=HTTPStatus.FOUND,
        )
        mock_duration.save.assert_called_once()
        mock_write_log.assert_called_once()

    @patch("afat.views.fatlinks.Duration.objects.get")
    def test_shows_error_when_fatlink_does_not_exist(self, mock_duration_get):
        """
        Test shows error when fatlink does not exist

        :param mock_duration_get:
        :type mock_duration_get:
        :return:
        :rtype:
        """

        mock_duration_get.side_effect = Duration.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_reopen_fatlink", args=["invalid_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))
        messages = list(get_messages(response.wsgi_request))
        self.assertTrue(
            any(
                "The hash you provided does not match with any FAT link." in str(m)
                for m in messages
            )
        )

    @patch("afat.views.fatlinks.Duration.objects.get")
    def test_shows_warning_when_fatlink_already_reopened(self, mock_duration_get):
        """
        Test shows warning when fatlink already reopened

        :param mock_duration_get:
        :type mock_duration_get:
        :return:
        :rtype:
        """

        mock_duration = MagicMock()
        mock_duration.fleet.reopened = True
        mock_duration_get.return_value = mock_duration

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.post(
            reverse("afat:fatlinks_reopen_fatlink", args=["test_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            reverse(
                "afat:fatlinks_details_fatlink",
                kwargs={"fatlink_hash": "test_hash"},
            ),
            target_status_code=HTTPStatus.FOUND,
        )
        messages = list(get_messages(response.wsgi_request))
        self.assertTrue(
            any("This FAT link has already been re-opened." in str(m) for m in messages)
        )

    @patch("afat.views.fatlinks.FatLink.objects.get")
    def test_closes_esi_fatlink_successfully(self, mock_get_fatlink):
        """
        Test closes esi fatlink successfully

        :param mock_get_fatlink:
        :type mock_get_fatlink:
        :return:
        :rtype:
        """

        mock_fatlink = MagicMock()
        mock_get_fatlink.return_value = mock_fatlink

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_close_esi_fatlink", args=["test_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))
        mock_fatlink.save.assert_called_once()
        self.assertFalse(mock_fatlink.is_registered_on_esi)

    @patch("afat.views.fatlinks.FatLink.objects.get")
    def test_handles_nonexistent_fatlink_gracefully(self, mock_get_fatlink):
        """
        Test handles nonexistent fatlink gracefully

        :param mock_get_fatlink:
        :type mock_get_fatlink:
        :return:
        :rtype:
        """

        mock_get_fatlink.side_effect = FatLink.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_close_esi_fatlink", args=["nonexistent_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))

    @patch("afat.views.fatlinks.FatLink.objects.get")
    @patch("afat.views.fatlinks.Fat.objects.get")
    @patch("afat.views.fatlinks.write_log")
    def test_deletes_fat_successfully(
        self, mock_write_log, mock_get_fat, mock_get_fatlink
    ):
        """
        Test deletes fat successfully

        :param mock_write_log:
        :type mock_write_log:
        :param mock_get_fat:
        :type mock_get_fat:
        :param mock_get_fatlink:
        :type mock_get_fatlink:
        :return:
        :rtype:
        """

        mock_fatlink = MagicMock(hash="test_hash")
        mock_fat = MagicMock(character=MagicMock(character_name="Test Character"))
        mock_get_fatlink.return_value = mock_fatlink
        mock_get_fat.return_value = mock_fat

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_delete_fat", args=["test_hash", 1])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            reverse("afat:fatlinks_details_fatlink", args=["test_hash"]),
            target_status_code=HTTPStatus.FOUND,
        )
        mock_fat.delete.assert_called_once()
        mock_write_log.assert_called_once_with(
            request=ANY,
            log_event=Log.Event.DELETE_FAT,
            log_text="The FAT for Test Character has been deleted",
            fatlink_hash="test_hash",
        )

    @patch("afat.views.fatlinks.FatLink.objects.get")
    def test_handles_delete_fat_with_nonexistent_fatlink_gracefully(
        self, mock_get_fatlink
    ):
        """
        Test handles delete fat with nonexistent fatlink gracefully

        :param mock_get_fatlink:
        :type mock_get_fatlink:
        :return:
        :rtype:
        """

        mock_get_fatlink.side_effect = FatLink.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_delete_fat", args=["invalid_hash", 1])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))

    @patch("afat.views.fatlinks.FatLink.objects.get")
    @patch("afat.views.fatlinks.Fat.objects.get")
    def test_handles_nonexistent_fat_gracefully(self, mock_get_fat, mock_get_fatlink):
        """
        Test handles nonexistent fat gracefully

        :param mock_get_fat:
        :type mock_get_fat:
        :param mock_get_fatlink:
        :type mock_get_fatlink:
        :return:
        :rtype:
        """

        mock_fatlink = MagicMock(hash="test_hash")
        mock_get_fatlink.return_value = mock_fatlink
        mock_get_fat.side_effect = Fat.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_delete_fat", args=["test_hash", 999])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))

    @patch("afat.views.fatlinks.FatLink.objects.get")
    @patch("afat.views.fatlinks.Fat.objects.filter")
    @patch("afat.views.fatlinks.write_log")
    def test_deletes_fatlink_and_associated_fats_successfully(
        self, mock_write_log, mock_fat_filter, mock_fatlink_get
    ):
        """
        Test deletes fatlink and associated fats successfully

        :param mock_write_log:
        :type mock_write_log:
        :param mock_fat_filter:
        :type mock_fat_filter:
        :param mock_fatlink_get:
        :type mock_fatlink_get:
        :return:
        :rtype:
        """

        mock_fatlink = MagicMock(hash="test_hash", pk=1)
        mock_fat_filter.return_value = MagicMock(delete=MagicMock())
        mock_fatlink_get.return_value = mock_fatlink

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_delete_fatlink", args=["test_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:fatlinks_overview"))
        mock_fat_filter.assert_called_once_with(fatlink_id=mock_fatlink.pk)
        mock_fat_filter.return_value.delete.assert_called_once()
        mock_fatlink.delete.assert_called_once()
        mock_write_log.assert_called_once_with(
            ANY,  # Allow any request object
            log_event=Log.Event.DELETE_FATLINK,
            log_text="FAT link deleted.",
            fatlink_hash="test_hash",
        )

    @patch("afat.views.fatlinks.FatLink.objects.get")
    def test_handles_delete_fatlink_with_nonexistent_fatlink_gracefully(
        self, mock_fatlink_get
    ):
        """
        Test handles delete fatlink with nonexistent fatlink gracefully

        :param mock_fatlink_get:
        :type mock_fatlink_get:
        :return:
        :rtype:
        """

        mock_fatlink_get.side_effect = FatLink.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_delete_fatlink", args=["invalid_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("afat:dashboard"))

    @patch("afat.views.fatlinks.Fat.objects.select_related_default")
    @patch("afat.views.fatlinks.convert_fats_to_dict")
    def test_returns_fats_for_valid_fatlink_hash(
        self, mock_convert_fats_to_dict, mock_select_related_default
    ):
        """
        Test returns fats for valid fatlink hash

        :param mock_convert_fats_to_dict:
        :type mock_convert_fats_to_dict:
        :param mock_select_related_default:
        :type mock_select_related_default:
        :return:
        :rtype:
        """

        # Create a mock Fat object with realistic attributes
        class MockFat:
            id = 1
            shiptype = "Omen"
            system = "Jita"

            class Character:
                character_name = "Test Character"

            character = Character()

            class FatLink:
                is_esilink = True
                is_registered_on_esi = True
                hash = "test_hash"
                created = datetime(2020, 4, 1, tzinfo=utc)
                fleet = "Test Fleet"
                doctrine = "Test Doctrine"
                fleet_type = "Test Fleet Type"

            fatlink = FatLink()

        mock_fat = MockFat()

        # Mock the queryset returned by select_related_default
        mock_queryset = MagicMock()
        mock_queryset.filter.return_value = [mock_fat]
        mock_select_related_default.return_value = mock_queryset

        # Mock the convert_fats_to_dict function to return a JSON-serializable dictionary
        mock_convert_fats_to_dict.return_value = {
            "id": mock_fat.id,
            "character": mock_fat.character.character_name,
            "shiptype": mock_fat.shiptype,
        }

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_ajax_get_fats_by_fatlink", args=["1231"])
        )

        # Assertions
        self.assertEqual(response.status_code, HTTPStatus.OK)
        mock_select_related_default.return_value.filter.assert_called_once_with(
            fatlink__hash="1231"
        )
        mock_convert_fats_to_dict.assert_called_once_with(request=ANY, fat=mock_fat)
        self.assertEqual(
            response.json(),
            [{"id": 1, "character": "Test Character", "shiptype": "Omen"}],
        )

    @patch("afat.views.fatlinks.Duration.objects.get")
    @patch("afat.views.fatlinks.FatLink.objects.select_related_default")
    def test_redirects_to_dashboard_when_invalid_hash_provided(
        self, mock_select_related_default, mock_duration_get
    ):
        """
        Test redirects to dashboard when invalid hash provided

        :param mock_select_related_default:
        :type mock_select_related_default:
        :param mock_duration_get:
        :type mock_duration_get:
        :return:
        :rtype:
        """

        mock_select_related_default.return_value.get.side_effect = FatLink.DoesNotExist

        self.client.force_login(self.user_with_manage_afat)
        response = self.client.get(
            reverse("afat:fatlinks_details_fatlink", args=["invalid_hash"])
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("afat:dashboard"))



# flake8: noqa

"""
Script to remove all generated FAT links from the database, which have been created by the script generate_fat_links.py.

This script can be executed directly from shell.
"""


# Standard Library
import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

# Django
import django

# init and setup django project
os.environ.setdefault(key="DJANGO_SETTINGS_MODULE", value="myauth.settings.local")
django.setup()

# Alliance Auth AFAT
from afat.models import FatLink

# Remove all generated FAT links
print("Removing all generated FAT links")
FatLink.objects.filter(fleet__startswith="Generated Fleet #").delete()
print("All generated FAT links have been removed")



# Generated by Django 3.1.1 on 2020-10-02 19:09

# Django
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("afat", "0009_auto_20200925_2206"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="afat",
            options={"verbose_name": "FAT", "verbose_name_plural": "FATs"},
        ),
        migrations.AlterModelOptions(
            name="afatdellog",
            options={
                "verbose_name": "Delete Log",
                "verbose_name_plural": "Delete Logs",
            },
        ),
        migrations.AlterModelOptions(
            name="afatlinktype",
            options={
                "verbose_name": "FAT Link Fleet Type",
                "verbose_name_plural": "FAT Link Fleet Types",
            },
        ),
        migrations.AlterModelOptions(
            name="clickafatduration",
            options={
                "verbose_name": "FAT Duration",
                "verbose_name_plural": "FAT Durations",
            },
        ),
        migrations.AlterModelOptions(
            name="manualafat",
            options={
                "verbose_name": "Manual FAT",
                "verbose_name_plural": "Manual FATs",
            },
        ),
    ]



"""
AFAT Smart Filter
"""

# Standard Library
import datetime
from collections import defaultdict

# Django
from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType

# Alliance Auth AFAT
from afat import __title__
from afat.models.afat import Fat, FleetType

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _get_threshold_date(timedelta_in_days: int) -> datetime.datetime:
    """
    Get the threshold date

    :param timedelta_in_days: The timedelta in days
    :type timedelta_in_days: int
    :return: The threshold date
    :rtype: datetime.datetime
    """

    return datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
        days=timedelta_in_days
    )


class BaseFilter(models.Model):
    """
    BaseFilter
    """

    name = models.CharField(
        max_length=500,
        help_text=_("The filter name that is shown to the admin."),
    )  # this is the name of the filter

    description = models.CharField(
        max_length=500,
        help_text=_("The filter description that is shown to end users."),
    )  # this is what is shown to the user

    class Meta:
        """
        Model meta definitions
        """

        abstract = True

    def __str__(self) -> str:
        """
        Model string representation

        :return: The model string representation
        :rtype: str
        """

        return f"{self.name}: {self.description}"

    def process_filter(self, user: User) -> bool:
        """
        Process the filter

        :param user: The user
        :type user: User
        :return: Return True when filter applies to the user, else False.
        :rtype: bool
        """

        raise NotImplementedError(_("Please create a filter!"))

    def audit_filter(self, users: models.QuerySet[User]) -> dict:
        """
        Return information for each given user weather they pass the filter,
        and a help message shown in the audit feature.

        :param users: The users
        :type users: models.QuerySet[User]
        :return: The audit information
        :rtype: dict
        """

        raise NotImplementedError(_("Please create an audit function!"))


class FatsInTimeFilter(BaseFilter):
    """
    FatsInTimeFilter
    """

    days = models.IntegerField(
        default=30, help_text=_("The number of days to look back for FATs.")
    )
    fats_needed = models.IntegerField(
        default=10, help_text=_("The number of FATs needed to pass the filter.")
    )
    fleet_types = models.ManyToManyField(
        to=FleetType,
        blank=True,
        help_text=_("Any of the selected fleet types are needed to pass the filter."),
    )
    ship_classes = models.ManyToManyField(
        to=EveType,
        blank=True,
        limit_choices_to={"eve_group__eve_category_id": 6},
        help_text=_("Any of the selected ship classes are needed to pass the filter."),
    )

    class Meta:
        """
        Model meta definitions
        """

        verbose_name = _("Smart Filter: FATs in time period")
        verbose_name_plural = verbose_name

    def process_filter(self, user: User):
        """
        Process the filter

        :param user:
        :type user:
        :return:
        :rtype:
        """

        try:
            start_time = _get_threshold_date(timedelta_in_days=self.days)
            character_list = user.character_ownerships.all().select_related("character")
            character_ids = character_list.values_list(
                "character__character_id", flat=True
            )
            ship_classes = self.ship_classes.all().values_list("name", flat=True)
            fleet_types = self.fleet_types.all().values_list("name", flat=True)

            fats = Fat.objects.filter(
                character__character_id__in=character_ids,
                fatlink__created__gte=start_time,
            )

            if ship_classes.exists():
                fats = fats.filter(shiptype__in=ship_classes)

            if fleet_types.exists():
                fats = fats.filter(fatlink__fleet_type__in=fleet_types)

            return fats.count() >= self.fats_needed
        except CharacterOwnership.DoesNotExist:
            # If the user does not have any characters, return False
            return False

    def audit_filter(self, users):
        """
        Audit the users for the filter

        :param users:
        :type users:
        :return:
        :rtype:
        """

        character_list = CharacterOwnership.objects.filter(user__in=users)
        start_time = _get_threshold_date(timedelta_in_days=self.days)
        ship_classes = self.ship_classes.all().values_list("name", flat=True)
        fleet_types = self.fleet_types.all().values_list("name", flat=True)

        fats = Fat.objects.filter(
            character__in=character_list.values("character"),
            fatlink__created__gte=start_time,
        ).select_related("character__character_ownership__user", "character")

        if ship_classes.exists():
            fats = fats.filter(shiptype__in=ship_classes)

        if fleet_types.exists():
            fats = fats.filter(fatlink__fleet_type__in=fleet_types)

        users = defaultdict(list)
        for f in fats:
            users[f.character.character_ownership.user.pk].append(f.id)

        output = defaultdict(lambda: {"message": 0, "check": False})
        for u, fat_list in users.items():
            pass_fail = False

            if len(fat_list) >= self.fats_needed:
                pass_fail = True
            output[u] = {"message": len(fat_list), "check": pass_fail}

        return output



"""
Our app setting
"""

# Standard Library
from re import RegexFlag

# Django
from django.apps import apps
from django.conf import settings


def fittings_installed() -> bool:
    """
    Check if the Fittings module is installed

    :return:
    :rtype:
    """

    return apps.is_installed(app_name="fittings")


def securegroups_installed() -> bool:
    """
    Check if the Alliance Auth Secure Groups module is installed

    :return:
    :rtype:
    """

    return apps.is_installed(app_name="securegroups")


def use_fittings_module_for_doctrines() -> bool:
    """
    Check if the Fittings module is used for doctrines

    :return:
    :rtype:
    """

    # Alliance Auth AFAT
    from afat.models import (  # pylint: disable=import-outside-toplevel, cyclic-import
        Setting,
    )

    return (
        fittings_installed() is True
        and Setting.get_setting(Setting.Field.USE_DOCTRINES_FROM_FITTINGS_MODULE)
        is True
    )


def debug_enabled() -> RegexFlag:
    """
    Check if DEBUG is enabled

    :return:
    :rtype:
    """

    return settings.DEBUG



"""
Logs related views
"""

# Django
from django.contrib.auth.decorators import login_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render

# Alliance Auth
from allianceauth.authentication.decorators import permissions_required
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.helper.views import convert_logs_to_dict
from afat.models import FatLink, Log, Setting

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.log_view"))
def overview(request: WSGIRequest) -> HttpResponse:
    """
    Logs view

    :param request:
    :type request:
    :return:
    :rtype:
    """

    logger.info(msg=f"Log view called by {request.user}")

    context = {"log_duration": Setting.get_setting(Setting.Field.DEFAULT_LOG_DURATION)}

    return render(
        request=request,
        template_name="afat/view/logs/logs-overview.html",
        context=context,
    )


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.log_view"))
def ajax_get_logs(
    request: WSGIRequest,  # pylint: disable=unused-argument
) -> JsonResponse:
    """
    Ajax call :: get all log entries

    :param request:
    :type request:
    :return:
    :rtype:
    """

    logs = Log.objects.select_related("user", "user__profile__main_character").all()
    fatlink_hashes = set(FatLink.objects.values_list("hash", flat=True))

    log_rows = [
        convert_logs_to_dict(log=log, fatlink_exists=log.fatlink_hash in fatlink_hashes)
        for log in logs
    ]

    return JsonResponse(data=log_rows, safe=False)



"""
Test access to the AFAT module
"""

# Standard Library
from http import HTTPStatus

# Django
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# Alliance Auth AFAT
from afat.tests import BaseTestCase
from afat.tests.fixtures.load_allianceauth import load_allianceauth

MODULE_PATH = "afat.views.statistics"


class TestAccesss(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()
        load_allianceauth()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)

        cls.user_without_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1002.character_id,
            permissions=["afat.basic_access"],
        )

    def test_should_show_afat_dashboard_for_user_with_basic_access(self):
        """
        Test should show afat dashboard for user with basic access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:dashboard")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_afat_dashboard_for_user_without_access(self):
        """
        Test should not show afat dashboard for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_without_access)

        url = reverse(viewname="afat:dashboard")
        res = self.client.get(path=url)

        self.assertNotEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=res.url, second="/account/login/?next=/fleet-activity-tracking/"
        )



"""
Statistics related views
"""

# Standard Library
import calendar
from collections import OrderedDict, defaultdict

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import Permission
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Count
from django.http import HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.datetime_safe import datetime
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_lazy

# Alliance Auth
from allianceauth.authentication.decorators import permissions_required
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.framework.api.evecharacter import get_main_character_from_evecharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.helper.views import (
    characters_with_permission,
    current_month_and_year,
    get_average_fats_by_corporations,
    get_fat_per_weekday,
    get_fats_per_hour,
    get_random_rgba_color,
    user_has_any_perms,
)
from afat.models import Fat
from afat.utils import get_or_create_alliance_info, get_or_create_corporation_info

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required()
@permission_required(perm="afat.basic_access")
def overview(request: WSGIRequest, year: int = None) -> HttpResponse:
    """
    Statistics main view

    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    if year is None:
        year = datetime.now().year

    user_can_see_other_corps = False

    if user_has_any_perms(
        user=request.user,
        perm_list=["afat.stats_corporation_other", "afat.manage_afat"],
    ):
        user_can_see_other_corps = True
        basic_access_permission = Permission.objects.select_related("content_type").get(
            content_type__app_label="afat", codename="basic_access"
        )

        characters_with_access = characters_with_permission(
            permission=basic_access_permission
        )

        data = {"No Alliance": [1]}
        sanity_check = {}

        # Group corporations by alliance
        for character_with_access in list(characters_with_access):
            alliance_name = character_with_access.alliance_name or "No Alliance"
            corp_id = character_with_access.corporation_id
            corp_name = character_with_access.corporation_name

            if alliance_name not in data:
                data[alliance_name] = [character_with_access.alliance_id]

            if corp_id not in sanity_check:
                data[alliance_name].append((corp_id, corp_name))
                sanity_check[corp_id] = corp_id
    elif request.user.has_perm(perm="afat.stats_corporation_own"):
        data = [
            (
                request.user.profile.main_character.corporation_id,
                request.user.profile.main_character.corporation_name,
            )
        ]
    else:
        data = None

    months = _calculate_year_stats(request=request, year=year)

    context = {
        "data": data,
        "charstats": months,
        "year": year,
        "year_current": datetime.now().year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
        "user_can_see_other_corps": user_can_see_other_corps,
    }

    logger.info(msg=f"Statistics overview called by {request.user}")

    return render(
        request=request,
        template_name="afat/view/statistics/statistics-overview.html",
        context=context,
    )


def _calculate_year_stats(request, year) -> dict:
    """
    Calculate statistics for the year

    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    months = {"total": {}, "characters": []}

    # Get all characters for the user and order by userprofile and character name
    characters = EveCharacter.objects.filter(
        character_ownership__user=request.user
    ).order_by("-userprofile", "character_name")

    # Get all FATs for the year and group by character and month
    fats_in_year = (
        Fat.objects.filter(fatlink__created__year=year, character__in=characters)
        .values("character__character_id", "fatlink__created__month")
        .annotate(fat_count=Count("id"))
    )

    # Initialize character data
    character_data = {
        char.character_id: {"name": char.character_name, "fats": {}}
        for char in characters
    }

    # Populate the months and character data
    for result in fats_in_year:
        month = int(result["fatlink__created__month"])
        char_id = int(result["character__character_id"])
        fat_count = int(result["fat_count"])

        # Update total fats per month
        if month not in months["total"]:
            months["total"][month] = 0

        months["total"][month] += fat_count

        # Update character fats per month
        character_data[char_id]["fats"][month] = fat_count

    # Sort character fats by month and add to the result,
    # excluding characters with no FATs
    for char_id, data in character_data.items():
        if data["fats"]:  # Only include characters with FATs
            sorted_fats = dict(sorted(data["fats"].items()))
            months["characters"].append((data["name"], sorted_fats, char_id))

    return months


@login_required()
@permission_required(perm="afat.basic_access")
def character(  # pylint: disable=too-many-locals
    request: WSGIRequest, charid: int, year: int = None, month: int = None
) -> HttpResponse:
    """
    Character statistics view

    :param request:
    :type request:
    :param charid:
    :type charid:
    :param year:
    :type year:
    :param month:
    :type month:
    :return:
    :rtype:
    """

    # Default to current year and month if not provided
    year = year or datetime.now().year
    month = month or datetime.now().month

    current_month, current_year = current_month_and_year()
    eve_character = EveCharacter.objects.get(character_id=charid)
    valid = [
        char.character for char in CharacterOwnership.objects.filter(user=request.user)
    ]

    can_view_character = True

    # Check if the user can view another corporation's statistics or manage AFAT
    if eve_character not in valid and not user_has_any_perms(
        user=request.user,
        perm_list=[
            "afat.stats_corporation_other",
            "afat.manage_afat",
        ],
    ):
        can_view_character = False

    # Check if the user is by any chance in the same corporation as the character
    # and can view own corporation statistics
    if (
        eve_character not in valid
        and eve_character.corporation_id
        == request.user.profile.main_character.corporation_id
        and request.user.has_perm(perm="afat.stats_corporation_own")
    ):
        can_view_character = True

    # If the user cannot view the character's statistics, send him home …
    if can_view_character is False:
        messages.warning(
            request=request,
            message=mark_safe(
                s=gettext(
                    "<h4>Warning!</h4>"
                    "<p>You do not have permission to view "
                    "statistics for this character.</p>"
                )
            ),
        )

        return redirect(to="afat:dashboard")

    fats = Fat.objects.filter(
        character__character_id=charid,
        fatlink__created__month=month,
        fatlink__created__year=year,
    )

    # Data for ship type pie chart
    data_ship_type = {}

    for fat in fats:
        if fat.shiptype in data_ship_type:
            continue

        data_ship_type[fat.shiptype] = fats.filter(shiptype=fat.shiptype).count()

    colors = []

    for _ in data_ship_type:
        bg_color_str = get_random_rgba_color()
        colors.append(bg_color_str)

    data_ship_type = [
        # Ship type can be None, so we need to convert to string here
        list(str(key) for key in data_ship_type),
        list(data_ship_type.values()),
        colors,
    ]

    # Data for by timeline Chart
    data_time = get_fats_per_hour(fats)

    context = {
        "character": eve_character,
        "month": month,
        "month_current": current_month,
        "month_prev": int(month) - 1,
        "month_next": int(month) + 1,
        "month_with_year": f"{year}{month:02d}",
        "month_current_with_year": f"{current_year}{current_month:02d}",
        "month_next_with_year": f"{year}{int(month) + 1:02d}",
        "month_prev_with_year": f"{year}{int(month) - 1:02d}",
        "year": year,
        "year_current": current_year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
        "data_ship_type": data_ship_type,
        "data_time": data_time,
        "fats": fats,
    }

    month_name = calendar.month_name[int(month)]
    logger.info(
        msg=(
            f"Character statistics for {eve_character} ({month_name} {year}) "
            f"called by {request.user}"
        )
    )

    return render(
        request=request,
        template_name="afat/view/statistics/statistics-character.html",
        context=context,
    )


@login_required()
@permissions_required(
    perm=(
        "afat.stats_corporation_other",
        "afat.stats_corporation_own",
        "afat.manage_afat",
    )
)
def ajax_get_monthly_fats_for_main_character(
    request: WSGIRequest,
    corporation_id: int,
    character_id: int,
    year: int,
    month: int,
) -> JsonResponse:
    """
    Ajax call :: Get monthly FATs for the main characters registered characters

    :param request: The request object
    :type request: WSGIRequest
    :param character_id: The main character
    :type character_id: EveCharacter
    :param year: The year
    :type year: int
    :param month: The month
    :type month: int
    :return: JSON response
    :rtype: JsonResponse
    """

    main_character = EveCharacter.objects.get(character_id=character_id)

    # Check character has permission to view other corps stats
    if int(request.user.profile.main_character.corporation_id) != int(
        main_character.corporation_id
    ) and not user_has_any_perms(
        user=request.user,
        perm_list=["afat.stats_corporation_other", "afat.manage_afat"],
    ):
        return JsonResponse(data="", safe=False)

    fats_per_character = (
        Fat.objects.filter(
            # corporation_eve_id=corporation_id,
            character__character_ownership__user=main_character.character_ownership.user,
            fatlink__created__month=month,
            fatlink__created__year=year,
        )
        .values(
            "corporation_eve_id",
            "character__character_id",
            "character__character_name",
            "character__corporation_id",
            "character__corporation_name",
        )
        .annotate(fat_count=Count("id"))
    )

    # if main_character.corporation_id != corporation_id:
    #     fats_per_character = fats_per_character.filter(
    #         character__corporation_id=corporation_id
    #     )

    info_button_text = gettext_lazy(
        "This character is in a different corporation and their FATs are "
        "not counted towards the statistics of the main corporation."
    )

    info_button = (
        "<sup>"
        f'<span class="ms-1 cursor-pointer"title="{info_button_text}" data-bs-tooltip="afat">'
        '<i class="fa-solid fa-circle-info"></i>'
        "</span>"
        "</sup>"
    )

    logger.debug("Fat per character: %s", fats_per_character)

    return JsonResponse(
        data=[
            {
                "character": item,
                "character_id": item["character__character_id"],
                "character_name": (
                    item["character__character_name"]
                    if item["corporation_eve_id"] == corporation_id
                    else f'{item["character__character_name"]} ({item["character__corporation_name"]}){info_button}'
                ),
                "fat_count": item["fat_count"],
                "show_details_button": (
                    f'<a class="btn btn-primary btn-sm" href="{reverse(viewname="afat:statistics_character", args=[item["character__character_id"], year, month])}"><i class="fa-solid fa-eye"></i></a>'
                ),
                "in_main_corp": item["corporation_eve_id"] == corporation_id,
            }
            for item in fats_per_character
        ],
        safe=False,
    )


@login_required()
@permissions_required(
    perm=(
        "afat.stats_corporation_other",
        "afat.stats_corporation_own",
        "afat.manage_afat",
    )
)
def corporation(  # pylint: disable=too-many-statements too-many-branches too-many-locals
    request: WSGIRequest, corpid: int = 0000, year: int = None, month: int = None
) -> HttpResponse:
    """
    Corp statistics view

    :param request:
    :type request:
    :param corpid:
    :type corpid:
    :param year:
    :type year:
    :param month:
    :type month:
    :return:
    :rtype:
    """

    # Default to current year if not provided
    year = year or datetime.now().year

    current_month, current_year = current_month_and_year()

    # Check character has permission to view other corps stats
    if int(request.user.profile.main_character.corporation_id) != int(
        corpid
    ) and not user_has_any_perms(
        user=request.user,
        perm_list=["afat.stats_corporation_other", "afat.manage_afat"],
    ):
        messages.warning(
            request=request,
            message=mark_safe(
                s=gettext(
                    "<h4>Warning!</h4>"
                    "<p>You do not have permission to view statistics "
                    "for that corporation.</p>"
                )
            ),
        )

        return redirect(to="afat:dashboard")

    corp = get_or_create_corporation_info(corporation_id=corpid)
    corp_name = corp.corporation_name

    if not month:
        fats_per_year = (
            Fat.objects.filter(
                corporation_eve_id=corpid,
                fatlink__created__year=year,
            )
            .values("fatlink__created__month")
            .annotate(fat_count=Count("id"))
        )

        months = []
        fats_per_year_total = 0

        for i in range(1, 13):
            corp_fats = next(
                (
                    item["fat_count"]
                    for item in fats_per_year
                    if item["fatlink__created__month"] == i
                ),
                0,
            )
            fats_per_year_total += corp_fats

            avg_fats = 0
            if corp.member_count > 0:
                avg_fats = corp_fats / corp.member_count

            if corp_fats > 0:
                months.append((i, corp_fats, round(avg_fats, 2)))

        context = {
            "corporation": corp_name,
            "months": months,
            "corpid": corpid,
            "year": year,
            "fats_per_year": fats_per_year_total,
            "year_current": current_year,
            "year_prev": int(year) - 1,
            "year_next": int(year) + 1,
            "type": 0,
        }

        return render(
            request=request,
            template_name="afat/view/statistics/statistics-corporation-year-overview.html",
            context=context,
        )

    fats = Fat.objects.filter(
        fatlink__created__month=month,
        fatlink__created__year=year,
        corporation_eve_id=corpid,
    ).select_related("character")

    # Data for Stacked Bar Graph
    # (label, color, [list of data for stack])
    data = defaultdict(lambda: defaultdict(int))
    character_ids = set()

    character_names_list = sorted(
        {fat.character.character_name for fat in fats}, key=str.lower
    )

    for fat in fats:
        data[fat.shiptype][fat.character.character_name] += 1
        character_ids.add(fat.character.character_id)

    data_stacked = [
        character_names_list,
        [
            (
                shiptype,
                get_random_rgba_color(),
                [counts[char] for char in character_names_list],
            )
            for shiptype, counts in data.items()
        ],
    ]

    # Data for By Time
    data_time = get_fats_per_hour(fats)

    # Data for By Weekday
    data_weekday = get_fat_per_weekday(fats)

    chars = {}
    main_chars = {}

    characters = EveCharacter.objects.filter(
        character_id__in=character_ids
    ).select_related("character_ownership__user")
    character_fat_counts = fats.values("character_id").annotate(fat_count=Count("id"))
    character_fat_map = {
        item["character_id"]: item["fat_count"] for item in list(character_fat_counts)
    }

    for char in characters:
        fat_c = character_fat_map.get(char.id, 0)
        chars[char.character_name] = (fat_c, char.character_id)
        main_character = get_main_character_from_evecharacter(character=char)

        if main_character and main_character.character_id not in main_chars:
            main_chars[main_character.character_id] = {
                "name": main_character.character_name,
                "id": main_character.character_id,
                "corporation_id": main_character.corporation_id,
                "corporation_name": main_character.corporation_name,
                "fats": fat_c,
            }
        elif main_character:
            main_chars[main_character.character_id]["fats"] += fat_c

    context = {
        "corp": corp,
        "corporation": corp_name,
        "month": month,
        "month_current": datetime.now().month,
        "month_prev": int(month) - 1,
        "month_next": int(month) + 1,
        "month_with_year": f"{year}{month:02d}",
        "month_current_with_year": f"{current_year}{current_month:02d}",
        "month_next_with_year": f"{year}{int(month) + 1:02d}",
        "month_prev_with_year": f"{year}{int(month) - 1:02d}",
        "year": year,
        "year_current": datetime.now().year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
        "data_stacked": data_stacked,
        "data_time": data_time,
        "data_weekday": data_weekday,
        "chars": chars,
        "main_chars": main_chars,
    }

    month_name = calendar.month_name[int(month)]

    logger.info(
        msg=(
            f"Corporation statistics for {corp_name} ({month_name} {year}) "
            f"called by {request.user}"
        )
    )

    return render(
        request=request,
        template_name="afat/view/statistics/statistics-corporation.html",
        context=context,
    )


@login_required()
@permissions_required(perm=("afat.stats_corporation_other", "afat.manage_afat"))
def alliance(  # pylint: disable=too-many-statements too-many-branches too-many-locals
    request: WSGIRequest, allianceid: int, year: int = None, month: int = None
) -> HttpResponse:
    """
    Alliance statistics view

    :param request:
    :type request:
    :param allianceid:
    :type allianceid:
    :param year:
    :type year:
    :param month:
    :type month:
    :return:
    :rtype:
    """

    # Default to current year if not provided
    year = year or datetime.now().year

    ally = (
        get_or_create_alliance_info(alliance_id=allianceid)
        if allianceid != "000"
        else None
    )
    alliance_name = ally.alliance_name if ally else "No Alliance"

    current_month, current_year = current_month_and_year()

    if not month:
        months = []

        ally_fats_by_month = (
            Fat.objects.filter(
                alliance_eve_id=allianceid,
                fatlink__created__year=year,
            )
            .order_by("fatlink__created__month")
            .values("fatlink__created__month")
            .annotate(fat_count=Count("id"))
        )

        for entry in ally_fats_by_month:
            months.append((entry["fatlink__created__month"], entry["fat_count"]))

        context = {
            "alliance": alliance_name,
            "months": months,
            "allianceid": allianceid,
            "year": year,
            "year_current": current_year,
            "year_prev": int(year) - 1,
            "year_next": int(year) + 1,
            "type": 1,
        }

        # /fleet-activity-tracking/statistics/alliance/<alliance_id>/<year>/
        return render(
            request=request,
            template_name="afat/view/statistics/statistics-alliance-year-overview.html",
            context=context,
        )

    fats = Fat.objects.filter(
        alliance_eve_id=allianceid,
        fatlink__created__month=month,
        fatlink__created__year=year,
    )

    # Data for ship type pie chart
    data_ship_type = fats.values("shiptype").annotate(count=Count("shiptype"))
    colors = [get_random_rgba_color() for _ in data_ship_type]

    data_ship_type = [
        [str(item["shiptype"]) for item in list(data_ship_type)],
        [item["count"] for item in list(data_ship_type)],
        colors,
    ]

    # Fats by corp and ship type
    data = {}
    corps_in_fats = set()

    for fat in fats:
        shiptype = fat.shiptype
        corp_name = fat.character.corporation_name

        if shiptype not in data:
            data[shiptype] = {}

        if corp_name not in data[shiptype]:
            data[shiptype][corp_name] = 0

        data[shiptype][corp_name] += 1
        corps_in_fats.add(corp_name)

    corps_in_fats = list(corps_in_fats)

    if None in data:
        data["Unknown"] = data.pop(None)

    data_stacked = [
        (key, get_random_rgba_color(), [value.get(corp, 0) for corp in corps_in_fats])
        for key, value in data.items()
    ]

    data_stacked = [corps_in_fats, data_stacked]

    corporations_in_alliance = EveCorporationInfo.objects.filter(alliance=ally)

    # Avg fats by corp
    data_avgs = get_average_fats_by_corporations(
        fats=fats, corporations=corporations_in_alliance
    )

    # Fats by Time
    data_time = get_fats_per_hour(fats=fats)

    # Fats by weekday
    data_weekday = get_fat_per_weekday(fats=fats)

    # Corp list
    corps = {}

    for corp in corporations_in_alliance:
        c_fats = fats.filter(corporation_eve_id=corp.corporation_id).count()
        avg = c_fats / corp.member_count
        corps[corp] = (corp.corporation_id, c_fats, round(avg, 2))

    corps = OrderedDict(sorted(corps.items(), key=lambda x: x[1][2], reverse=True))

    context = {
        "alliance": alliance_name,
        "ally": ally,
        "month": month,
        "month_current": current_month,
        "month_prev": int(month) - 1,
        "month_next": int(month) + 1,
        "month_with_year": f"{year}{month:02d}",
        "month_current_with_year": f"{current_year}{current_month:02d}",
        "month_next_with_year": f"{year}{int(month) + 1:02d}",
        "month_prev_with_year": f"{year}{int(month) - 1:02d}",
        "year": year,
        "year_current": current_year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
        "data_stacked": data_stacked,
        "data_avgs": data_avgs,
        "data_time": data_time,
        "data_weekday": data_weekday,
        "corps": corps,
        "data_ship_type": data_ship_type,
    }

    month_name = calendar.month_name[int(month)]
    logger.info(
        msg=(
            f"Alliance statistics for {alliance_name} ({month_name} {year}) "
            f"called by {request.user}"
        )
    )

    return render(
        request=request,
        template_name="afat/view/statistics/statistics-alliance.html",
        context=context,
    )



"""
Tasks
"""

# Standard Library
from datetime import timedelta

# Third Party
import kombu
from bravado.exception import HTTPNotFound
from celery import group, shared_task

# Django
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.esi import fetch_esi_status
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.models import Fat, FatLink, Log, Setting
from afat.providers import esi
from afat.utils import get_or_create_character

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)

ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_MAX_RETRIES = 3
ESI_MAX_ERROR_COUNT = 3
ESI_ERROR_GRACE_TIME = 75

TASK_TIME_LIMIT = 120  # Stop after 2 minutes

# Params for all tasks
TASK_DEFAULT_KWARGS = {"time_limit": TASK_TIME_LIMIT, "max_retries": ESI_MAX_RETRIES}


@shared_task(**{**TASK_DEFAULT_KWARGS}, **{"base": QueueOnce})
def process_fats(data_list, data_source: str, fatlink_hash: str):
    """
    Process FAT link data

    :param data_list:
    :type data_list:
    :param data_source:
    :type data_source:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    logger.debug(f"Data Source: {data_source}")

    if data_source == "esi":
        logger.info(
            msg=(
                f'Valid fleet for FAT link hash "{fatlink_hash}" found '
                "registered via ESI, checking for new pilots"
            )
        )

        my_tasks = [
            process_character.si(
                character_id=char["character_id"],
                solar_system_id=char["solar_system_id"],
                ship_type_id=char["ship_type_id"],
                fatlink_hash=fatlink_hash,
            )
            for char in list(data_list)
        ]

        if my_tasks:
            logger.debug(
                "Creating group of tasks for %s characters and adding them to the queue",
                len(my_tasks),
            )

            try:
                group(my_tasks).delay()
            except kombu.exceptions.EncodeError:
                logger.debug(
                    msg=(
                        "No changes to the current fleet rooster, nothing to add to the queue"
                    )
                )


@shared_task
def process_character(
    character_id: int, solar_system_id: int, ship_type_id: int, fatlink_hash: str
) -> None:
    """
    Process character

    :param character_id:
    :param solar_system_id:
    :param ship_type_id:
    :param fatlink_hash:
    :return:
    """

    character = get_or_create_character(character_id=character_id)
    link = FatLink.objects.get(hash=fatlink_hash)

    solar_system_name = esi.client.Universe.get_universe_systems_system_id(
        system_id=solar_system_id
    ).result()["name"]
    ship_name = esi.client.Universe.get_universe_types_type_id(
        type_id=ship_type_id
    ).result()["name"]

    fat, created = Fat.objects.get_or_create(
        fatlink=link,
        character=character,
        corporation_eve_id=character.corporation_id,
        alliance_eve_id=character.alliance_id,
        defaults={"system": solar_system_name, "shiptype": ship_name},
    )

    if created:
        logger.info(
            f"New Pilot: Adding {character} in {solar_system_name} flying "
            f'a {ship_name} to FAT link "{fatlink_hash}" (FAT ID {fat.pk})'
        )
    else:
        logger.debug(
            f"Pilot {character} already registered for FAT link {fatlink_hash} "
            f"with FAT ID {fat.pk}"
        )


def _close_esi_fleet(fatlink: FatLink, reason: str) -> None:
    """
    Closing ESI fleet

    :param fatlink:
    :type fatlink:
    :param reason:
    :type reason:
    :return:
    :rtype:
    """

    logger.info(
        msg=f'Closing ESI FAT link with hash "{fatlink.hash}". Reason: {reason}'
    )

    fatlink.is_registered_on_esi = False
    fatlink.save()


def _esi_fatlinks_error_handling(error_key: str, fatlink: FatLink) -> None:
    """
    ESI error handling

    :param error_key:
    :type error_key:
    :param fatlink:
    :type fatlink:
    :return:
    :rtype:
    """

    time_now = timezone.now()
    grace_period = time_now - timedelta(seconds=ESI_ERROR_GRACE_TIME)

    if (
        fatlink.last_esi_error == error_key
        and fatlink.last_esi_error_time >= grace_period
        and fatlink.esi_error_count >= ESI_MAX_ERROR_COUNT
    ):
        _close_esi_fleet(fatlink=fatlink, reason=error_key.label)

        return

    fatlink.esi_error_count = (
        fatlink.esi_error_count + 1
        if fatlink.last_esi_error == error_key
        and fatlink.last_esi_error_time >= grace_period
        else 1
    )
    fatlink.last_esi_error = error_key
    fatlink.last_esi_error_time = time_now
    fatlink.save()

    logger.info(
        msg=(
            f'FAT link "{fatlink.hash}" Error: "{error_key.label}" '
            f"({fatlink.esi_error_count} of {ESI_MAX_ERROR_COUNT})."
        )
    )


def _check_for_esi_fleet(fatlink: FatLink) -> dict | None:
    """
    Check if there is a fleet for this FAT link registered on ESI

    :param fatlink:
    :type fatlink:
    :return:
    :rtype:
    """

    required_scopes = ["esi-fleets.read_fleet.v1"]
    fleet_commander_id = fatlink.character.character_id

    try:
        esi_token = Token.get_token(
            character_id=fleet_commander_id, scopes=required_scopes
        )
        fleet_from_esi = esi.client.Fleets.get_characters_character_id_fleet(
            character_id=fleet_commander_id,
            token=esi_token.valid_access_token(),
        ).result()

        logger.debug("Fleet from ESI: %s", fleet_from_esi)
        logger.debug("FAT Link ESI fleet ID: %s", fatlink.esi_fleet_id)

        if not fleet_from_esi or fatlink.esi_fleet_id != fleet_from_esi["fleet_id"]:
            raise HTTPNotFound

        return {"fleet": fleet_from_esi, "token": esi_token}
    except HTTPNotFound:
        error_key = FatLink.EsiError.NOT_IN_FLEET
    except Exception:  # pylint: disable=broad-exception-caught
        error_key = FatLink.EsiError.NO_FLEET

    _esi_fatlinks_error_handling(error_key=error_key, fatlink=fatlink)

    return None


@shared_task()
def _process_esi_fatlink(fatlink: FatLink) -> None:
    """
    Processing ESI FAT link

    :param fatlink:
    :type fatlink:
    :return:
    :rtype:
    """

    logger.info(msg=f'Processing ESI FAT link with hash "{fatlink.hash}"')

    if not fatlink.creator.profile.main_character:
        _close_esi_fleet(fatlink=fatlink, reason="No FAT link creator available.")

        return

    # Check if there is a fleet
    esi_fleet = _check_for_esi_fleet(fatlink=fatlink)
    logger.debug("ESI fleet: %s", esi_fleet)

    if not esi_fleet:
        return

    # Check if we deal with the fleet boss here
    try:
        esi_fleet_member = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=esi_fleet["fleet"]["fleet_id"],
            token=esi_fleet["token"].valid_access_token(),
        ).result()
    except Exception:  # pylint: disable=broad-exception-caught
        _esi_fatlinks_error_handling(
            error_key=FatLink.EsiError.NOT_FLEETBOSS, fatlink=fatlink
        )

        return

    # Process fleet members
    logger.debug(
        msg=f'Processing fleet members for ESI FAT link with hash "{fatlink.hash}"'
    )

    process_fats.delay(
        data_list=esi_fleet_member,
        data_source="esi",
        fatlink_hash=fatlink.hash,
    )


@shared_task(**{**TASK_DEFAULT_KWARGS, "base": QueueOnce})
def update_esi_fatlinks() -> None:
    """
    Checking ESI fat links for changes
    """

    esi_status = fetch_esi_status()

    # Abort if ESI seems offline or above the error limit
    if not esi_status.is_ok:
        logger.warning("ESI doesn't seem to be available at this time. Aborting.")

        return

    esi_fatlinks = (
        FatLink.objects.select_related_default()
        .filter(is_esilink=True, is_registered_on_esi=True)
        .distinct()
    )

    logger.debug(msg=f"Found {len(esi_fatlinks)} ESI FAT links to process")
    logger.debug("ESI FAT Links: %s", esi_fatlinks)

    for fatlink in esi_fatlinks:
        _process_esi_fatlink(fatlink=fatlink)


@shared_task
def logrotate():
    """
    Remove logs older than AFAT_DEFAULT_LOG_DURATION

    :return:
    :rtype:
    """

    logger.info(
        msg=f"Cleaning up logs older than {Setting.get_setting(Setting.Field.DEFAULT_LOG_DURATION)} days"
    )

    Log.objects.filter(
        log_time__lte=timezone.now()
        - timedelta(days=Setting.get_setting(Setting.Field.DEFAULT_LOG_DURATION))
    ).delete()



"""
Test checks for installed modules we might use
"""

# Alliance Auth AFAT
from afat.models import Setting
from afat.tests import BaseTestCase


class TestSettingModel(BaseTestCase):
    """
    Test the Setting model
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup the test class.

        :return:
        :rtype:
        """

        super().setUpClass()

    def test_default_fatlink_expiry_time(self):
        """
        Test default fatlink expiry time.

        :return:
        :rtype:
        """

        self.assertEqual(
            first=Setting.get_setting(Setting.Field.DEFAULT_FATLINK_EXPIRY_TIME),
            second=60,
        )

    def test_get_setting_with_invalid_key(self):
        """
        Test get_setting with an invalid key.

        :return:
        :rtype:
        """

        with self.assertRaises(expected_exception=KeyError):
            Setting.get_setting("invalid_key")

    def test_get_setting_with_valid_key(self):
        """
        Test get_setting with a valid key.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()
        settings.default_fatlink_expiry_time = 75
        settings.save()

        self.assertEqual(
            first=Setting.get_setting(Setting.Field.DEFAULT_FATLINK_EXPIRY_TIME),
            second=75,
        )

    def test_model_str(self):
        """
        Test model string representation.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()

        self.assertEqual(first=str(settings), second="AFAT Settings")



"""
Statistics URLs
These URLs are used for static pages and are prefixed with `static/`
"""

# Django
from django.urls import path

# Alliance Auth AFAT
from afat.views import statistics

urls = [
    # Stats main page
    path(
        route="",
        view=statistics.overview,
        name="statistics_overview",
    ),
    path(
        route="<int:year>/",
        view=statistics.overview,
        name="statistics_overview",
    ),
    # Stats corp
    path(
        route="corporation/",
        view=statistics.corporation,
        name="statistics_corporation",
    ),
    path(
        route="corporation/<int:corpid>/",
        view=statistics.corporation,
        name="statistics_corporation",
    ),
    path(
        route="corporation/<int:corpid>/<int:year>/",
        view=statistics.corporation,
        name="statistics_corporation",
    ),
    path(
        route="corporation/<int:corpid>/<int:year>/<int:month>/",
        view=statistics.corporation,
        name="statistics_corporation",
    ),
    # Stats char
    path(
        route="character/",
        view=statistics.character,
        name="statistics_character",
    ),
    path(
        route="character/<int:charid>/",
        view=statistics.character,
        name="statistics_character",
    ),
    path(
        route="character/<int:charid>/<int:year>/<int:month>/",
        view=statistics.character,
        name="statistics_character",
    ),
    # Stats alliance
    path(
        route="alliance/",
        view=statistics.alliance,
        name="statistics_alliance",
    ),
    path(
        route="alliance/<int:allianceid>/",
        view=statistics.alliance,
        name="statistics_alliance",
    ),
    path(
        route="alliance/<int:allianceid>/<int:year>/",
        view=statistics.alliance,
        name="statistics_alliance",
    ),
    path(
        route="alliance/<int:allianceid>/<int:year>/<int:month>/",
        view=statistics.alliance,
        name="statistics_alliance",
    ),
]
